from typing import TYPE_CHECKING, List, Optional

import torch
import torch.utils.data.dataloader
import tqdm

from llmcompressor.modifiers.utils.hooks import HooksMixin
from llmcompressor.pipelines.cache import IntermediatesCache
from llmcompressor.pipelines.layer_sequential.helpers import (
    capture_first_layer_intermediates,
    match_modules,
    maybe_inject_pos_embeddings,
    to_next_layer_kwargs,
)
from llmcompressor.utils.helpers import DisableQuantization, calibration_forward_context

if TYPE_CHECKING:
    from llmcompressor.modifiers import Modifier

__all__ = ["run_pipeline"]


def run_pipeline(
    model: torch.nn.Module,
    dataloader: torch.utils.data.DataLoader,
    sequential_targets: List[str],
    callback_modifier: Optional["Modifier"] = None,
):
    """
    Run a layer-wise sequential data pipeline according to the following steps:

    1. Layers are identified according to `sequential_targets`
    2. A hook is attached to the first layer. This hook raises an exception which is
        then caught and used to capture the input arguments to the first layer
    3. The inputs to the first layer are used to calibrate the first layer, and the
        output of the previous layer is used as inputs to calibrate the next layer

    This pipeline requires that the model have distinct layers defined in its
    architecture and that the outputs of the previous layer are exactly the inputs
    to the next layer. This is violated by encoder-decoder architectures among others.

    If your model architecture violates these assumptions, consider using the sequential
    pipeline (see llmcompressor.pipelines.sequential). Architectures which are known to
    fail these assumptions include GPT-J and most vision language models

    :param model: model being calibrated
    :param dataloader: loads data for calibration
    :param sequential_targets: patterns which match to the layer modules of the model
    :param callback_modifier: Temporary HACK which should be replaced by event callback
    """
    # find layers
    layers = match_modules(model, sequential_targets)

    with calibration_forward_context(model), DisableQuantization(model):
        # prepare intermediates cache
        intermediates: IntermediatesCache = capture_first_layer_intermediates(
            model, layers[0], dataloader
        )

        num_layers = len(layers)
        for layer_index, layer in enumerate(layers):
            # prepare tqdm description texts
            calib_desc = f"({layer_index + 1}/{num_layers}): Calibrating"
            prop_desc = f"({layer_index + 1}/{num_layers}): Propagating"

            # do an preliminary pass to trigger modifier hooks
            for batch_index in tqdm.tqdm(range(len(dataloader)), desc=calib_desc):
                inputs = intermediates.fetch(batch_index)
                layer(**inputs)

            # TODO: replace with a lifecycle event
            if callback_modifier:
                callback_modifier.on_sequential_batch_end()

            # this pass does not trigger modifier hooks
            # and is only used for capturing outputs from the newly compressed modules
            with HooksMixin.disable_hooks():
                for batch_index in tqdm.tqdm(range(len(dataloader)), desc=prop_desc):
                    inputs = intermediates.fetch(batch_index)
                    output = layer(**inputs)

                    if layer_index < num_layers - 1:
                        next_layer = layers[layer_index + 1]
                        output = to_next_layer_kwargs(output, next_layer)
                        output = maybe_inject_pos_embeddings(output, next_layer, inputs)

                        intermediates.delete(batch_index)
                        intermediates.update(batch_index, output)



from typing import List, Optional, Union

import torch
from torch.nn.modules.linear import Identity
from torch.quantization import QuantWrapper

__all__ = ["get_leaf_operations", "is_quantized", "get_precision_information"]


def get_leaf_operations(
    model: torch.nn.Module,
    operations_to_skip: Optional[List[torch.nn.Module]] = None,
    operations_to_unwrap: Optional[List[torch.nn.Module]] = None,
) -> List[torch.nn.Module]:
    """
    Get the leaf operations in the model
    (those that do not have operations as children)

    :param model: the model to get the leaf operations from
    :param operations_to_skip: a list of leaf operations that will be
        omitted when getting the leaf operations. If None passed, by
        default the Identity operation will be skipped
    :param operations_to_unwrap: a list of operations that will be unwrapped
        when getting the leaf operations. Unwrapping means that we directly
        add the module(s) that is/are wrapped by the operation (i.e. operation's
        `module` attribute) to the list
        of leaf operations. If None passed, by default the QuantWrapper
        operation will be unwrapped
    :return: a list of the leaf operations
    """
    if operations_to_skip is None:
        operations_to_skip = [Identity]

    if operations_to_unwrap is None:
        operations_to_unwrap = [QuantWrapper]

    leaf_operations = []
    children = list(model.children())

    if children == []:
        return model
    else:
        for child in children:
            if isinstance(child, tuple(operations_to_unwrap)):
                leaf_operations.append(child.module)
                continue
            try:
                leaf_operations.extend(get_leaf_operations(child))
            except TypeError:
                leaf_operations.append(get_leaf_operations(child))
    leaf_operations = [
        op for op in leaf_operations if not isinstance(op, tuple(operations_to_skip))
    ]
    return leaf_operations


def is_quantized(operation: torch.nn.Module) -> bool:
    """
    Check whether the operation is quantized (contains
    a quantization scheme)
    """
    return hasattr(operation, "quantization_scheme")


def get_precision_information(
    operation: torch.nn.Module,
) -> Union[None, int, "QuantizationScheme"]:  # noqa F821
    """
    Get the information about the precision of the operation.

    1)  If operation is quantized, returns the quantization
        scheme of the operation.
    2)  If operation is not quantized, returns the numer of bits
        of the operation's weights.
    3)  If operation is not quantized and does not have a weights,
        returns None.

    :param operation: the operation to get the quantization scheme from
    :return: the quantization scheme of the operation, the number of bits
        of the operation's weights, or None if the operation is not quantized
        and does not have a weight
    """

    if hasattr(operation, "quantization_scheme"):
        return getattr(operation, "quantization_scheme")
    elif hasattr(operation, "weight"):
        return _get_num_bits(operation.weight.dtype)
    else:
        return None


def _get_num_bits(dtype: torch.dtype) -> int:
    # Get the number of bits of a torch dtype
    if dtype == torch.float16 or dtype == torch.bfloat16:
        return 16
    elif dtype == torch.float32:
        return 32
    elif dtype == torch.float64:
        return 64
    elif dtype == torch.int8:
        return 8
    elif dtype == torch.int16:
        return 16
    elif dtype == torch.int32:
        return 32
    elif dtype == torch.int64:
        return 64
    else:
        raise ValueError("Unknown dtype: {}".format(dtype))



import threading
from contextlib import contextmanager
from typing import Any, Optional

from llmcompressor.core.events import EventType
from llmcompressor.core.session import CompressionSession
from llmcompressor.core.state import ModifiedState

__all__ = [
    "create_session",
    "active_session",
    "reset_session",
    "callbacks",
    "LifecycleCallbacks",
]


_global_session = CompressionSession()
_local_storage = threading.local()
_local_storage.session = _global_session


@contextmanager
def create_session() -> CompressionSession:
    """
    Context manager to create and yield a new session for sparsification.
    This will set the active session to the new session for the duration
    of the context.

    :return: the new session
    """
    global _local_storage
    orig_session = getattr(_local_storage, "session", None)
    new_session = CompressionSession()
    _local_storage.session = new_session
    try:
        yield new_session
    finally:
        _local_storage.session = orig_session


def active_session() -> CompressionSession:
    """
    :return: the active session for sparsification
    """
    global _local_storage
    return getattr(_local_storage, "session", _global_session)


def reset_session():
    """
    Reset the currently active session to its initial state
    """
    session = active_session()
    session._lifecycle.reset()


class LifecycleCallbacks:
    """
    A class for invoking lifecycle events for the active session
    """

    @classmethod
    def event(cls, event_type: EventType, **kwargs) -> ModifiedState:
        """
        Invoke an event for the active session

        :param event_type: the event type to invoke
        :param kwargs: additional kwargs to pass to the current session's event method
        :return: the modified state of the active session after invoking the event
        """
        if event_type in [EventType.INITIALIZE, EventType.FINALIZE]:
            raise ValueError(
                f"Cannot invoke {event_type} event. "
                f"Use the corresponding method instead."
            )

        # skip event callbacks if no recipe was provided
        if not active_session().lifecycle.recipe_container.check_any_recipe_exists():
            return

        return active_session().event(event_type, **kwargs)

    @classmethod
    def batch_start(cls, batch_data: Optional[Any] = None, **kwargs) -> ModifiedState:
        """
        Invoke a batch start event for the active session

        :param batch_data: the batch data to use for the event
        :param kwargs: additional kwargs to pass to the current session's event method
        :return: the modified state of the active session after invoking the event
        """
        return cls.event(EventType.BATCH_START, batch_data=batch_data, **kwargs)

    @classmethod
    def loss_calculated(cls, loss: Optional[Any] = None, **kwargs) -> ModifiedState:
        """
        Invoke a loss calculated event for the active session

        :param loss: the loss to use for the event
        :param kwargs: additional kwargs to pass to the current session's event method
        :return: the modified state of the active session after invoking the event
        """
        # log loss if loss calculated
        active_session()._log_loss(event_type=EventType.LOSS_CALCULATED, loss=loss)
        return cls.event(EventType.LOSS_CALCULATED, loss=loss, **kwargs)

    @classmethod
    def optim_pre_step(cls, **kwargs) -> ModifiedState:
        """
        Invoke an optimizer pre-step event for the active session

        :param kwargs: additional kwargs to pass to the current session's event method
        :return: the modified state of the active session after invoking the event
        """
        return cls.event(EventType.OPTIM_PRE_STEP, **kwargs)

    @classmethod
    def optim_post_step(cls, **kwargs) -> ModifiedState:
        """
        Invoke an optimizer post-step event for the active session

        :param kwargs: additional kwargs to pass to the current session's event method
        :return: the modified state of the active session after invoking the event
        """
        return cls.event(EventType.OPTIM_POST_STEP, **kwargs)

    @classmethod
    def batch_end(cls, **kwargs) -> ModifiedState:
        """
        Invoke a batch end event for the active session

        :param kwargs: additional kwargs to pass to the current session's event method
        :return: the modified state of the active session after invoking the event
        """
        active_session()._log_model_info()
        return cls.event(EventType.BATCH_END, **kwargs)


callbacks = LifecycleCallbacks



import os
import shutil
import unittest

import pytest
from compressed_tensors.compressors import ModelCompressor
from parameterized import parameterized_class
from transformers import AutoConfig

from tests.testing_utils import parse_params, requires_gpu

CONFIGS_DIRECTORY = "tests/llmcompressor/transformers/finetune/finetune_oneshot_configs"
GPU_CONFIGS_DIRECTORY = (
    "tests/llmcompressor/transformers/finetune/finetune_oneshot_configs/gpu"
)


class TestOneshotAndFinetune(unittest.TestCase):
    def _test_oneshot_and_finetune(self):
        from llmcompressor import oneshot, train

        splits = {"train": "train[:5%]", "calibration": "train[5%:10%]"}
        if self.dataset == "ultrachat-200k":
            splits = {"train": "train_gen[:5%]", "calibration": "train_gen[5%:10%]"}

        oneshot_args = dict(
            dataset=self.dataset,
            splits=splits,
            recipe=self.recipe,
            num_calibration_samples=64,
            oneshot_device=self.device,
            dataset_config_name=self.dataset_config_name,
            concatenate_data=self.concat_txt,
            output_dir=self.output,
        )

        train_args = dict(
            num_train_epochs=self.num_train_epochs,
            precision="bfloat16",
            bf16=True,
        )
        oneshot_model = oneshot(
            model=self.model,
            **oneshot_args,
            stage="test_oneshot_stage",
        )

        train(
            model=oneshot_model,
            **oneshot_args,
            **train_args,
            stage="test_train_stage",
        )

        config_sparse_applied = ModelCompressor.parse_sparsity_config(
            AutoConfig.from_pretrained(
                os.path.join(self.output, "test_oneshot_stage")
            ).quantization_config
        )
        config_finetune_applied = ModelCompressor.parse_sparsity_config(
            AutoConfig.from_pretrained(
                os.path.join(self.output, "test_train_stage")
            ).quantization_config
        )
        # model is first sparsified, then finetuned, both should have the same sparsity
        assert config_sparse_applied["global_sparsity"] == pytest.approx(
            config_finetune_applied["global_sparsity"], abs=1e-5
        )

    def tearDown(self):
        # TODO: we get really nice stats from finetune that we should log
        # stored in results.json
        if os.path.isdir(self.output):
            shutil.rmtree(self.output)


@pytest.mark.integration
@parameterized_class(parse_params(CONFIGS_DIRECTORY))
class TestOneshotAndFinetuneSmall(TestOneshotAndFinetune):
    model = None
    dataset = None
    recipe = None
    dataset_config_name = None
    num_train_epochs = None
    concat_txt = None

    def setUp(self):
        import torch

        self.device = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.output = "./finetune_output"

    def test_oneshot_then_finetune_small(self):
        self._test_oneshot_and_finetune()


@requires_gpu
@pytest.mark.integration
@parameterized_class(parse_params(GPU_CONFIGS_DIRECTORY))
class TestOneshotAndFinetuneGPU(TestOneshotAndFinetune):
    model = None
    dataset = None
    recipe = None
    dataset_config_name = None
    num_train_epochs = None
    concat_txt = None

    def setUp(self):
        import torch
        from transformers import AutoModelForCausalLM

        self.device = "cuda:0"
        self.output = "./finetune_output"

        self.model = AutoModelForCausalLM.from_pretrained(
            self.model, device_map=self.device, torch_dtype=torch.bfloat16
        )

    def test_oneshot_then_finetune_gpu(self):
        self._test_oneshot_and_finetune()



from typing import Union

from datasets import Dataset, DatasetDict, IterableDataset
from transformers import (
    BaseImageProcessor,
    FeatureExtractionMixin,
    PreTrainedTokenizer,
    ProcessorMixin,
)

# Tokenizer or Processor. Processors do not inherit from a unified base class
Processor = Union[
    PreTrainedTokenizer, BaseImageProcessor, FeatureExtractionMixin, ProcessorMixin
]

# Supported dataset types, IterableDataset is a streamed dataset
DatasetType = Union[Dataset, DatasetDict, IterableDataset]



import os
from typing import Dict, Iterable

import pytest
import torch
from torch import Tensor
from torch.nn import Linear, Module, ReLU, Sequential

from llmcompressor.pytorch.utils import (
    sanitize_kwargs_for_module,
    tensor_forward_with_input_args,
    tensor_sparsity,
    tensors_module_forward,
    tensors_to_device,
    tensors_to_precision,
)


@pytest.mark.unit
@pytest.mark.skipif(
    os.getenv("NM_ML_SKIP_PYTORCH_TESTS", False),
    reason="Skipping pytorch tests",
)
@pytest.mark.parametrize(
    "tensors",
    [
        (),
        [],
        {},
        torch.randn(1, 8, 16, 32),
        torch.randn(8, 8, 16, 32),
        (torch.randn(1, 8), torch.randn(8, 8)),
        [torch.randn(1, 8), torch.randn(8, 8)],
        {"key": torch.randn(1, 8), "key2": torch.randn(8, 8)},
        [[torch.randn(1, 8)], torch.randn(8, 8)],
    ],
)
def test_tensors_to_device_cpu(tensors):
    out = tensors_to_device(tensors, "cpu")

    if isinstance(out, Tensor):
        assert not out.is_cuda
    elif isinstance(out, Iterable):
        for tens in out:
            if isinstance(tens, Tensor):
                assert not tens.is_cuda
    elif isinstance(out, Dict):
        for key, tens in out.items():
            if isinstance(tens, Tensor):
                assert not tens.is_cuda


@pytest.mark.unit
@pytest.mark.skipif(
    os.getenv("NM_ML_SKIP_PYTORCH_TESTS", False),
    reason="Skipping pytorch tests",
)
@pytest.mark.parametrize(
    "tensors",
    [
        (),
        [],
        {},
        torch.randn(1, 8, 16, 32),
        torch.randn(8, 8, 16, 32),
        (torch.randn(1, 8), torch.randn(8, 8)),
        [torch.randn(1, 8), torch.randn(8, 8)],
        {"key": torch.randn(1, 8), "key2": torch.randn(8, 8)},
        [[torch.randn(1, 8)], torch.randn(8, 8)],
    ],
)
@pytest.mark.skipif(not torch.cuda.is_available(), reason="requires cuda availability")
def test_tensors_to_device_cuda(tensors):
    out = tensors_to_device(tensors, "cuda")

    if isinstance(out, Tensor):
        assert out.is_cuda
    elif isinstance(out, Iterable):
        for tens in out:
            if isinstance(tens, Tensor):
                assert tens.is_cuda
    elif isinstance(out, Dict):
        for key, tens in out.items():
            if isinstance(tens, Tensor):
                assert tens.is_cuda


@pytest.mark.unit
@pytest.mark.skipif(
    os.getenv("NM_ML_SKIP_PYTORCH_TESTS", False),
    reason="Skipping pytorch tests",
)
@pytest.mark.parametrize(
    "tensors",
    [
        (),
        [],
        {},
        torch.randn(1, 8, 16, 32),
        torch.randn(8, 8, 16, 32),
        (torch.randn(1, 8), torch.randn(8, 8)),
        [torch.randn(1, 8), torch.randn(8, 8)],
        {"key": torch.randn(1, 8), "key2": torch.randn(8, 8)},
        [[torch.randn(1, 8)], torch.randn(8, 8)],
    ],
)
def test_tensors_to_precision_full_cpu(tensors):
    out = tensors_to_precision(tensors, True)

    if isinstance(out, Tensor):
        assert out.dtype == torch.float32
    elif isinstance(out, Iterable):
        for tens in out:
            if isinstance(tens, Tensor):
                assert tens.dtype == torch.float32
    elif isinstance(out, Dict):
        for key, tens in out.items():
            if isinstance(tens, Tensor):
                assert tens.dtype == torch.float32


@pytest.mark.unit
@pytest.mark.skipif(
    os.getenv("NM_ML_SKIP_PYTORCH_TESTS", False),
    reason="Skipping pytorch tests",
)
@pytest.mark.parametrize(
    "tensors",
    [
        (),
        [],
        {},
        torch.randn(1, 8, 16, 32),
        torch.randn(8, 8, 16, 32),
        (torch.randn(1, 8), torch.randn(8, 8)),
        [torch.randn(1, 8), torch.randn(8, 8)],
        {"key": torch.randn(1, 8), "key2": torch.randn(8, 8)},
        [[torch.randn(1, 8)], torch.randn(8, 8)],
    ],
)
def test_tensors_to_precision_half_cpu(tensors):
    out = tensors_to_precision(tensors, False)

    if isinstance(out, Tensor):
        assert out.dtype == torch.float16
    elif isinstance(out, Iterable):
        for tens in out:
            if isinstance(tens, Tensor):
                assert tens.dtype == torch.float16
    elif isinstance(out, Dict):
        for key, tens in out.items():
            if isinstance(tens, Tensor):
                assert tens.dtype == torch.float16


@pytest.mark.unit
@pytest.mark.skipif(
    os.getenv("NM_ML_SKIP_PYTORCH_TESTS", False),
    reason="Skipping pytorch tests",
)
@pytest.mark.parametrize(
    "tensors",
    [
        (),
        [],
        {},
        torch.randn(1, 8, 16, 32),
        torch.randn(8, 8, 16, 32),
        (torch.randn(1, 8), torch.randn(8, 8)),
        [torch.randn(1, 8), torch.randn(8, 8)],
        {"key": torch.randn(1, 8), "key2": torch.randn(8, 8)},
        [[torch.randn(1, 8)], torch.randn(8, 8)],
    ],
)
@pytest.mark.skipif(not torch.cuda.is_available(), reason="requires cuda availability")
def test_tensors_to_precision_full_cuda(tensors):
    tensors = tensors_to_device(tensors, "cuda")
    out = tensors_to_precision(tensors, True)

    if isinstance(out, Tensor):
        assert out.dtype == torch.float32
    elif isinstance(out, Iterable):
        for tens in out:
            if isinstance(tens, Tensor):
                assert tens.dtype == torch.float32
    elif isinstance(out, Dict):
        for key, tens in out.items():
            if isinstance(tens, Tensor):
                assert tens.dtype == torch.float32


@pytest.mark.unit
@pytest.mark.skipif(
    os.getenv("NM_ML_SKIP_PYTORCH_TESTS", False),
    reason="Skipping pytorch tests",
)
@pytest.mark.parametrize(
    "tensors",
    [
        (),
        [],
        {},
        torch.randn(1, 8, 16, 32),
        torch.randn(8, 8, 16, 32),
        (torch.randn(1, 8), torch.randn(8, 8)),
        [torch.randn(1, 8), torch.randn(8, 8)],
        {"key": torch.randn(1, 8), "key2": torch.randn(8, 8)},
        [[torch.randn(1, 8)], torch.randn(8, 8)],
    ],
)
@pytest.mark.skipif(not torch.cuda.is_available(), reason="requires cuda availability")
def test_tensors_to_precision_half_cuda(tensors):
    tensors = tensors_to_device(tensors, "cuda")
    out = tensors_to_precision(tensors, False)

    if isinstance(out, Tensor):
        assert out.dtype == torch.float16
    elif isinstance(out, Iterable):
        for tens in out:
            if isinstance(tens, Tensor):
                assert tens.dtype == torch.float16
    elif isinstance(out, Dict):
        for key, tens in out.items():
            if isinstance(tens, Tensor):
                assert tens.dtype == torch.float16


@pytest.mark.unit
@pytest.mark.skipif(
    os.getenv("NM_ML_SKIP_PYTORCH_TESTS", False),
    reason="Skipping pytorch tests",
)
class SimpleModule(Module):
    def __init__(self, input_size: int):
        super().__init__()
        self.fc1 = Linear(input_size, 16, bias=True)
        self.relu1 = ReLU()
        self.fc2 = Linear(16, 32, bias=True)
        self.relu2 = ReLU()

    def forward(self, inp):
        out = self.fc1(inp)
        out = self.relu1(out)
        out = self.fc2(out)
        out = self.relu2(out)

        return out

    @staticmethod
    def example_input(batch_size: int, input_size: int):
        return torch.randn(batch_size, input_size)

    @staticmethod
    def example_output(batch_size: int):
        return torch.randn(batch_size, 32)


class ComplexModule(Module):
    def __init__(self, input_size_one: int, input_size_two: int):
        super().__init__()
        self.branch1 = Sequential(
            Linear(input_size_one, 16, bias=True), ReLU(), Linear(16, 32), ReLU()
        )
        self.branch2 = Sequential(
            Linear(input_size_two, 16, bias=True),
            ReLU(),
            Linear(16, 32, bias=True),
            ReLU(),
        )
        self.tower = Sequential(Linear(64, 32, bias=True), ReLU())

    def forward(self, inp_one, inp_two):
        out_one = self.branch1(inp_one)
        out_two = self.branch2(inp_two)
        out = torch.cat([out_one, out_two], dim=1)
        out = self.tower(out)

        return out

    @staticmethod
    def example_list_input(batch_size: int, input_size_one: int, input_size_two: int):
        return [
            torch.randn(batch_size, input_size_one),
            torch.randn(batch_size, input_size_two),
        ]

    @staticmethod
    def example_dict_input(batch_size: int, input_size_one: int, input_size_two: int):
        return {
            "inp_one": torch.randn(batch_size, input_size_one),
            "inp_two": torch.randn(batch_size, input_size_two),
        }

    @staticmethod
    def example_output(batch_size: int):
        return torch.randn(batch_size, 32)


@pytest.mark.unit
@pytest.mark.skipif(
    os.getenv("NM_ML_SKIP_PYTORCH_TESTS", False),
    reason="Skipping pytorch tests",
)
@pytest.mark.parametrize(
    "module,tensors,check_feat_lab_inp",
    [
        (SimpleModule(8), SimpleModule.example_input(1, 8), False),
        (SimpleModule(8), SimpleModule.example_input(16, 8), False),
        (ComplexModule(8, 4), ComplexModule.example_list_input(1, 8, 4), False),
        (ComplexModule(8, 4), ComplexModule.example_list_input(16, 8, 4), False),
        (ComplexModule(8, 4), ComplexModule.example_dict_input(1, 8, 4), False),
        (ComplexModule(8, 4), ComplexModule.example_dict_input(16, 8, 4), False),
        (
            SimpleModule(8),
            (SimpleModule.example_input(1, 8), SimpleModule.example_output(1)),
            True,
        ),
        (
            SimpleModule(8),
            [SimpleModule.example_input(16, 8), SimpleModule.example_output(16)],
            True,
        ),
        (
            ComplexModule(8, 4),
            [
                ComplexModule.example_list_input(1, 8, 4),
                ComplexModule.example_output(1),
            ],
            True,
        ),
        (
            ComplexModule(8, 4),
            (
                ComplexModule.example_list_input(16, 8, 4),
                ComplexModule.example_output(16),
            ),
            True,
        ),
        (
            ComplexModule(8, 4),
            (
                ComplexModule.example_dict_input(1, 8, 4),
                ComplexModule.example_output(1),
            ),
            True,
        ),
        (
            ComplexModule(8, 4),
            [
                ComplexModule.example_dict_input(16, 8, 4),
                ComplexModule.example_output(16),
            ],
            True,
        ),
    ],
)
def test_tensors_module_forward(module, tensors, check_feat_lab_inp):
    out = tensors_module_forward(tensors, module, check_feat_lab_inp)
    assert len(out)


@pytest.mark.unit
@pytest.mark.skipif(
    os.getenv("NM_ML_SKIP_PYTORCH_TESTS", False),
    reason="Skipping pytorch tests",
)
@pytest.mark.parametrize(
    "module,tensors,check_feat_lab_inp",
    [
        (SimpleModule(8), SimpleModule.example_input(1, 8), False),
        (SimpleModule(8), SimpleModule.example_input(16, 8), False),
        (ComplexModule(8, 4), ComplexModule.example_list_input(1, 8, 4), False),
        (ComplexModule(8, 4), ComplexModule.example_list_input(16, 8, 4), False),
        (ComplexModule(8, 4), ComplexModule.example_dict_input(1, 8, 4), False),
        (ComplexModule(8, 4), ComplexModule.example_dict_input(16, 8, 4), False),
        (
            SimpleModule(8),
            (SimpleModule.example_input(1, 8), SimpleModule.example_output(1)),
            True,
        ),
        (
            SimpleModule(8),
            [SimpleModule.example_input(16, 8), SimpleModule.example_output(16)],
            True,
        ),
        (
            ComplexModule(8, 4),
            [
                ComplexModule.example_list_input(1, 8, 4),
                ComplexModule.example_output(1),
            ],
            True,
        ),
        (
            ComplexModule(8, 4),
            (
                ComplexModule.example_list_input(16, 8, 4),
                ComplexModule.example_output(16),
            ),
            True,
        ),
        (
            ComplexModule(8, 4),
            (
                ComplexModule.example_dict_input(1, 8, 4),
                ComplexModule.example_output(1),
            ),
            True,
        ),
        (
            ComplexModule(8, 4),
            [
                ComplexModule.example_dict_input(16, 8, 4),
                ComplexModule.example_output(16),
            ],
            True,
        ),
    ],
)
@pytest.mark.skipif(not torch.cuda.is_available(), reason="requires cuda availability")
def test_tensors_module_forward_cuda(module, tensors, check_feat_lab_inp):
    module = module.to("cuda")
    tensors = tensors_to_device(tensors, "cuda")
    out = tensors_module_forward(tensors, module, check_feat_lab_inp)
    assert out is not None


@pytest.mark.unit
@pytest.mark.flaky(reruns=2, min_passes=1)
@pytest.mark.skipif(
    os.getenv("NM_ML_SKIP_PYTORCH_TESTS", False),
    reason="Skipping pytorch tests",
)
@pytest.mark.parametrize(
    "tensor,dim,expected_sparsity",
    [
        (torch.zeros(8, 16), None, torch.tensor(1.0)),
        (torch.zeros(8, 16), 0, torch.ones(8)),
        (torch.zeros(8, 16), 1, torch.ones(16)),
        (torch.zeros(8, 16), [0, 1], torch.ones(8, 16)),
        (torch.zeros(8, 16), [1, 0], torch.ones(16, 8)),
        (torch.zeros(8, 16, 32, 8), [3, 1, 2], torch.ones(8, 16, 32)),
        (torch.ones(8, 16), None, torch.tensor(0.0)),
        (torch.ones(8, 16), 0, torch.zeros(8)),
        (torch.ones(8, 16), 1, torch.zeros(16)),
        (torch.ones(8, 16), [0, 1], torch.zeros(8, 16)),
        (torch.ones(8, 16), [1, 0], torch.zeros(16, 8)),
        (torch.ones(8, 16, 32, 8), [3, 1, 2], torch.zeros(8, 16, 32)),
        (torch.randn(8, 16), None, torch.tensor(0.0)),
        (torch.randn(8, 16), 0, torch.zeros(8)),
        (torch.randn(8, 16), 1, torch.zeros(16)),
        (torch.randn(8, 16), [0, 1], torch.zeros(8, 16)),
        (torch.randn(8, 16), [1, 0], torch.zeros(16, 8)),
        (torch.randn(8, 16, 32, 8), [3, 1, 2], torch.zeros(8, 16, 32)),
        (
            torch.tensor([10.0, 0.0, 1.0, 3.0, 2.0, 0.0, 8.0, 0.0, 5.0, 0.0]),
            None,
            torch.tensor(0.4),
        ),
    ],
)
def test_tensor_sparsity(tensor, dim, expected_sparsity):
    sparsity = tensor_sparsity(tensor, dim)
    assert expected_sparsity.shape == sparsity.shape
    assert torch.sum((sparsity - expected_sparsity).abs()) < 0.001


@pytest.mark.unit
@pytest.mark.flaky(reruns=2, min_passes=1)
@pytest.mark.skipif(
    os.getenv("NM_ML_SKIP_PYTORCH_TESTS", False),
    reason="Skipping pytorch tests",
)
@pytest.mark.parametrize(
    "tensor,dim,expected_sparsity",
    [
        (torch.zeros(8, 16), None, torch.tensor(1.0)),
        (torch.zeros(8, 16), 0, torch.ones(8)),
        (torch.zeros(8, 16, 32, 8), [3, 1, 2], torch.ones(8, 16, 32)),
        (torch.ones(8, 16), None, torch.tensor(0.0)),
        (torch.ones(8, 16), 0, torch.zeros(8)),
        (torch.ones(8, 16, 32, 8), [3, 1, 2], torch.zeros(8, 16, 32)),
        (torch.randn(8, 16), None, torch.tensor(0.0)),
        (torch.randn(8, 16), 0, torch.zeros(8)),
        (torch.randn(8, 16, 32, 8), [3, 1, 2], torch.zeros(8, 16, 32)),
        (
            torch.tensor([10.0, 0.0, 1.0, 3.0, 2.0, 0.0, 8.0, 0.0, 5.0, 0.0]),
            None,
            torch.tensor(0.4),
        ),
    ],
)
@pytest.mark.skipif(not torch.cuda.is_available(), reason="requires cuda availability")
def test_tensor_sparsity_cuda(tensor, dim, expected_sparsity):
    tensor = tensor.to("cuda")
    sparsity = tensor_sparsity(tensor, dim)
    assert expected_sparsity.shape == sparsity.shape
    assert torch.sum((sparsity.detach().cpu() - expected_sparsity).abs()) < 0.001


class TestSanitizeKwargsForModule:
    @pytest.fixture
    def module(self):
        return Linear(10, 20)

    def test_sanitize_kwargs_for_module_not_dict(self, module):
        # Test with kwargs that are not a dictionary
        with pytest.raises(TypeError):
            sanitize_kwargs_for_module("not a dictionary", module)

    def test_sanitize_kwargs_for_module_not_in_signature(self, module):
        # Test with kwargs that are not in the signature of the forward method
        kwargs = {"not_in_signature": 123}
        sanitized_kwargs = sanitize_kwargs_for_module(kwargs, module)
        assert sanitized_kwargs == {}

    def test_sanitize_kwargs_for_module_in_signature(self, module):
        # Test with kwargs that are in the signature of the forward method
        kwargs = {"input": torch.randn(1, 10)}
        sanitized_kwargs = sanitize_kwargs_for_module(kwargs, module)
        assert sanitized_kwargs == kwargs


class TestTensorForwardWithInputArgs:
    @pytest.fixture
    def module(self):
        return Linear(10, 20)

    def test_tensor_forward_with_input_args(self, module):
        # Test with valid inputs and input_kwargs
        inputs = torch.randn(1, 10)
        input_kwargs = {}
        output = tensor_forward_with_input_args(module, inputs, input_kwargs)
        assert output.shape == (1, 20)

        # Test with input_kwargs that are not in the signature of the forward method
        input_kwargs = {"not_in_signature": 123}
        tensor_forward_with_input_args(module, inputs, input_kwargs)



from dataclasses import dataclass, field
from typing import List, Optional

from llmcompressor.modifiers import Modifier
from llmcompressor.recipe.recipe import (
    Recipe,
    RecipeArgsInput,
    RecipeInput,
    RecipeStageInput,
    RecipeTuple,
)

__all__ = ["RecipeContainer"]


@dataclass
class RecipeContainer:
    """
    A container for recipes to be used in a session. Provides utilities
    to update the recipes and compile them into a single recipe.

    :param compiled_recipe: the compiled recipe from the recipes list
    :param recipes: the list of RecipeTuple instances to be compiled
    :param applied_stages: list of recipe stages that have already been applied
    """

    compiled_recipe: Optional[Recipe] = None
    recipes: List[RecipeTuple] = field(default_factory=list)
    applied_stages: List[str] = field(default_factory=list)

    def prepend(
        self,
        recipe: Optional[RecipeInput] = None,
        recipe_stage: Optional[RecipeStageInput] = None,
        recipe_args: Optional[RecipeArgsInput] = None,
    ):
        recipe_tuples = self._prepare_tuples(recipe, recipe_stage, recipe_args)
        self.recipes = recipe_tuples + self.recipes
        self._check_compile_recipe()

    def append(
        self,
        recipe: Optional[RecipeInput] = None,
        recipe_stage: Optional[RecipeStageInput] = None,
        recipe_args: Optional[RecipeArgsInput] = None,
    ):
        recipe_tuples = self._prepare_tuples(recipe, recipe_stage, recipe_args)
        self.recipes = self.recipes + recipe_tuples
        self._check_compile_recipe()

    def get_modifiers(self) -> List[Modifier]:
        if self.compiled_recipe is None:
            return []

        return self.compiled_recipe.create_modifier()

    def _prepare_tuples(
        self,
        recipe: Optional[RecipeInput] = None,
        recipe_stage: Optional[RecipeStageInput] = None,
        recipe_args: Optional[RecipeArgsInput] = None,
    ) -> List[RecipeTuple]:
        if recipe is None or (isinstance(recipe, list) and len(recipe) == 0):
            return []

        # prepare recipe
        if isinstance(recipe, Modifier) or (
            isinstance(recipe, list)
            and all(isinstance(mod, Modifier) for mod in recipe)
        ):
            recipe = Recipe.create_instance(recipe)

        if not isinstance(recipe, list):
            recipe = [recipe]

        recipe = [
            Recipe.create_instance(rec) if isinstance(rec, str) else rec
            for rec in recipe
        ]

        # prepare stage
        if recipe_stage is None:
            recipe_stage = [None] * len(recipe)
        else:
            if not isinstance(recipe_stage, list):
                recipe_stage = [[recipe_stage]] * len(recipe)
            if not isinstance(recipe_stage[0], list):
                recipe_stage = [recipe_stage] * len(recipe)

        # prepare args
        if recipe_args is None:
            recipe_args = [{}] * len(recipe)
        elif not isinstance(recipe_args, list):
            recipe_args = [recipe_args] * len(recipe)

        # validation
        if len(recipe) != len(recipe_stage) or len(recipe) != len(recipe_args):
            raise ValueError(
                "recipe, recipe_stage, and recipe_args must be the same length"
            )

        # create tuples
        return [
            RecipeTuple(rec, stage, args)
            for rec, stage, args in zip(recipe, recipe_stage, recipe_args)
        ]

    def update_applied_stages(self, new_stages: List[str]):
        """
        Updates the applied_stages list with new stages, indicating their structure
        has already been applied

        :param new_stages: new stage names to add
        """
        for stage in new_stages:
            if stage not in self.applied_stages:
                self.applied_stages.append(stage)

    def _check_compile_recipe(self):
        """
        Check if the recipes need to be compiled into a single recipe and
        compile them if they do.

        :return: True if the recipes were compiled, False otherwise
        """
        if self.recipes:
            self.compiled_recipe = Recipe.simplify_combine_recipes(self.recipes)

    def check_any_recipe_exists(self) -> bool:
        """
        Checks if any recipes have been added to the container, compiled or not

        :return: True if any recipes exist in the container, False otherwise
        """
        if self.compiled_recipe is not None:
            return True
        if len(self.recipes) > 0:
            return True

        return False



import re
from typing import Callable, Dict, Sequence, Tuple, Union

import torch
import torch.nn.functional as TF
from torch import Tensor
from torch.nn import Module

from llmcompressor.core import State

__all__ = [
    "TensorOrCollectionType",
    "ProjectionFuncType",
    "CreateProjectionFuncType",
    "TransformFuncType",
    "CreateTransformFuncType",
    "ComparisonFuncType",
    "CreateComparisonFuncType",
    "KDFactory",
    "recursive_apply",
    "recursive_combine",
    "identity_transform",
    "softmax_transform",
    "log_softmax_transform",
    "normalize_transform",
    "l1_comparison",
    "l2_comparison",
    "inner_product_comparison",
    "cosine_similarity_comparison",
    "kl_divergence_comparison",
    "cross_entropy_comparison",
]


TensorOrCollectionType = Union[Tensor, Sequence[Tensor], Dict[str, Tensor]]
ProjectionFuncType = Callable[
    [TensorOrCollectionType, TensorOrCollectionType], TensorOrCollectionType
]
CreateProjectionFuncType = Callable[
    [str, Module, Module, State], Tuple[ProjectionFuncType, ProjectionFuncType]
]
TransformFuncType = Callable[[TensorOrCollectionType], TensorOrCollectionType]
CreateTransformFuncType = Callable[[str, Module, Module, State], TransformFuncType]
ComparisonFuncType = Callable[
    [TensorOrCollectionType, TensorOrCollectionType], TensorOrCollectionType
]
CreateComparisonFuncType = Callable[[str, Module, Module, State], ComparisonFuncType]


class KDFactory:
    registry_projections: Dict[str, CreateProjectionFuncType] = {}
    registry_transforms: Dict[str, CreateTransformFuncType] = {}
    registry_comparisons: Dict[str, CreateComparisonFuncType] = {}

    @staticmethod
    def register_projection(name: str, func: CreateProjectionFuncType):
        KDFactory.registry_projections[name] = func

    @staticmethod
    def register_projection_decorator(name: str):
        def inner(func: CreateProjectionFuncType):
            KDFactory.registry_projections[name] = func
            return func

        return inner

    @staticmethod
    def create_projection(
        name: str, student_layer: Module, teacher_layer: Module, state: State, **kwargs
    ) -> Tuple[ProjectionFuncType, ProjectionFuncType]:
        for pattern, creator in KDFactory.registry_projections:
            match = pattern == name

            if not match:
                try:
                    match = re.match(pattern, name)
                except Exception:
                    pass

            if match:
                return creator(
                    name=name,
                    student_layer=student_layer,
                    teacher_layer=teacher_layer,
                    state=state,
                    **kwargs,
                )

        raise ValueError(f"Invalid projection name: {name}")

    @staticmethod
    def register_transform(name: str, func: CreateTransformFuncType):
        KDFactory.registry_transforms[name] = func

    @staticmethod
    def register_transform_decorator(name: str):
        def inner(func: CreateTransformFuncType):
            KDFactory.registry_transforms[name] = func
            return func

        return inner

    @staticmethod
    def create_transform(
        name: str,
        layer: Module,
        state: State,
        **kwargs,
    ) -> TransformFuncType:
        for pattern, creator in KDFactory.registry_transforms.items():
            match = pattern == name

            if not match:
                try:
                    match = re.match(pattern, name)
                except Exception:
                    pass

            if match:
                return creator(
                    name=name,
                    layer=layer,
                    state=state,
                    **kwargs,
                )

        raise ValueError(f"Invalid transform name: {name}")

    @staticmethod
    def register_comparison(name: str, func):
        KDFactory.registry_comparisons[name] = func

    @staticmethod
    def register_comparison_decorator(name: str):
        def inner(func):
            KDFactory.registry_comparisons[name] = func
            return func

        return inner

    @staticmethod
    def create_comparison(
        name: str, student_layer: Module, teacher_layer: Module, state: State, **kwargs
    ) -> ComparisonFuncType:
        for pattern, creator in KDFactory.registry_comparisons.items():
            match = pattern == name

            if not match:
                try:
                    match = re.match(pattern, name)
                except Exception:
                    pass

            if match:
                return creator(
                    name=name,
                    student_layer=student_layer,
                    teacher_layer=teacher_layer,
                    state=state,
                    **kwargs,
                )

        raise ValueError(f"Invalid comparison name: {name}")


def recursive_apply(
    val: TensorOrCollectionType,
    func: Callable[[Tensor], Tensor],
) -> TensorOrCollectionType:
    if isinstance(val, Tensor):
        return func(val)

    if isinstance(val, Sequence):
        return [recursive_apply(item, func) for item in val]

    if isinstance(val, dict):
        return {key: recursive_apply(item, func) for key, item in val.items()}

    raise ValueError(f"Unsupported type for recursive_apply: {type(val)}")


def recursive_combine(
    val_one: TensorOrCollectionType,
    val_two: TensorOrCollectionType,
    func: Callable[[Tensor, Tensor], Tensor],
):
    if not isinstance(val_one, type(val_two)):
        raise ValueError(
            f"val_one type of {type(val_one)} must match "
            f"val_two type of {type(val_two)}"
        )

    if isinstance(val_one, Tensor):
        return func(val_one, val_two)

    if isinstance(val_one, Sequence):
        return [
            recursive_combine(item_one, item_two, func)
            for item_one, item_two in zip(val_one, val_two)
        ]

    if isinstance(val_one, dict):
        return {
            key: recursive_combine(val_one[key], val_two[key], func)
            for key in val_one.keys()
        }

    raise ValueError(f"Unsupported type for recursive_combine: {type(val_one)}")


@KDFactory.register_transform_decorator("identity")
def identity_transform(name: str, **kwargs):
    if name != "identity":
        raise ValueError(f"Invalid transform name: {name}")

    def _create_transform(val: TensorOrCollectionType) -> TensorOrCollectionType:
        return val

    return _create_transform


@KDFactory.register_transform_decorator("softmax")
def softmax_transform(name: str, temperature: float = 1.0, dim: int = -1, **kwargs):
    if name != "softmax":
        raise ValueError(f"Invalid transform name: {name}")

    def _softmax(val: Tensor) -> Tensor:
        val = val / temperature

        return torch.softmax(val, dim=dim)

    def _create_transform(val: TensorOrCollectionType) -> TensorOrCollectionType:
        return recursive_apply(val, _softmax)

    return _create_transform


@KDFactory.register_transform_decorator("log_softmax")
def log_softmax_transform(name: str, temperature: float = 1.0, dim: int = -1, **kwargs):
    if name != "log_softmax":
        raise ValueError(f"Invalid transform name: {name}")

    def _log_softmax(val: Tensor) -> Tensor:
        val = val / temperature

        return torch.log_softmax(val, dim=dim)

    def _create_transform(val: TensorOrCollectionType) -> TensorOrCollectionType:
        return recursive_apply(val, _log_softmax)

    return _create_transform


@KDFactory.register_transform_decorator("normalize")
def normalize_transform(
    name: str,
    p: float = 1,
    dim: int = -1,
    eps: float = 1e-12,
    mean: bool = False,
    std: bool = False,
    **kwargs,
):
    if name != "normalize":
        raise ValueError(f"Invalid transform name: {name}")

    def _normalize(val: Tensor) -> Tensor:
        out = TF.normalize(val, p=p, dim=dim, eps=eps)

        if mean:
            out = out - out.mean(dim=dim, keepdim=True)

        if std:
            out = out / out.std(dim=dim, keepdim=True)

        return out

    def _create_transform(val: TensorOrCollectionType) -> TensorOrCollectionType:
        return recursive_apply(val, _normalize)

    return _create_transform


@KDFactory.register_comparison_decorator("l1_distance")
def l1_comparison(name: str, dim: int = -1, **kwargs):
    if name != "l1_distance":
        raise ValueError(f"Invalid comparison name: {name}")

    def _l1(val_one: Tensor, val_two: Tensor) -> Tensor:
        return torch.sum(torch.abs(val_one - val_two), dim=dim)

    def _create_comparison(
        val_one: TensorOrCollectionType, val_two: TensorOrCollectionType
    ) -> TensorOrCollectionType:
        return recursive_combine(val_one, val_two, _l1)

    return _create_comparison


@KDFactory.register_comparison_decorator("l2_distance")
def l2_comparison(name: str, dim: int = -1, **kwargs):
    if name != "l2_distance":
        raise ValueError(f"Invalid comparison name: {name}")

    def _l2(val_one: Tensor, val_two: Tensor) -> Tensor:
        return torch.sum((val_one - val_two) ** 2, dim=dim)

    def _create_comparison(
        val_one: TensorOrCollectionType, val_two: TensorOrCollectionType
    ) -> TensorOrCollectionType:
        return recursive_combine(val_one, val_two, _l2)

    return _create_comparison


@KDFactory.register_comparison_decorator("inner_product")
def inner_product_comparison(name: str, dim: int = -1, **kwargs):
    if name != "inner_product":
        raise ValueError(f"Invalid comparison name: {name}")

    def _inner_product(val_one: Tensor, val_two: Tensor) -> Tensor:
        return torch.sum(val_one * val_two, dim=dim)

    def _create_comparison(
        val_one: TensorOrCollectionType, val_two: TensorOrCollectionType
    ) -> TensorOrCollectionType:
        return recursive_combine(val_one, val_two, _inner_product)

    return _create_comparison


@KDFactory.register_comparison_decorator("cosine_similarity")
def cosine_similarity_comparison(name: str, dim: int = -1, **kwargs):
    if name != "cosine_similarity":
        raise ValueError(f"Invalid comparison name: {name}")

    def _cosine_similarity(val_one: Tensor, val_two: Tensor) -> Tensor:
        return torch.sum(val_one * val_two, dim=dim) / (
            torch.norm(val_one, dim=dim) * torch.norm(val_two, dim=dim)
        )

    def _create_comparison(
        val_one: TensorOrCollectionType, val_two: TensorOrCollectionType
    ) -> TensorOrCollectionType:
        return recursive_combine(val_one, val_two, _cosine_similarity)

    return _create_comparison


@KDFactory.register_comparison_decorator("kl_divergence")
def kl_divergence_comparison(
    name: str, dim: int = -1, temperature: float = 1.0, **kwargs
):
    if name != "kl_divergence":
        raise ValueError(f"Invalid comparison name: {name}")

    def _kl_divergence(val_one: Tensor, val_two: Tensor) -> Tensor:
        val_one = val_one / temperature
        val_two = val_two / temperature

        return torch.sum(val_one * torch.log(val_one / val_two), dim=dim)

    def _create_comparison(
        val_one: TensorOrCollectionType, val_two: TensorOrCollectionType
    ) -> TensorOrCollectionType:
        return recursive_combine(val_one, val_two, _kl_divergence)

    return _create_comparison


@KDFactory.register_comparison_decorator("cross_entropy")
def cross_entropy_comparison(
    name: str, temperature: float = 1.0, reduction: str = "none", **kwargs
):
    if name != "cross_entropy":
        raise ValueError(f"Invalid projection name: {name}")

    def _cross_entropy(val_one: Tensor, val_two: Tensor) -> Tensor:
        val_one = val_one / temperature
        val_two = val_two / temperature

        return TF.cross_entropy(val_one, val_two, reduction=reduction)

    def _create_comparison(
        val_one: TensorOrCollectionType, val_two: TensorOrCollectionType
    ) -> TensorOrCollectionType:
        return recursive_combine(val_one, val_two, _cross_entropy)

    return _create_comparison


@KDFactory.register_comparison_decorator("square_head")
def square_head_comparison(name: str, **kwargs):
    if name != "square_head":
        raise ValueError(f"Invalid projection name: {name}")

    def _square_head(val_one: Tensor, val_two: Tensor) -> Tensor:
        numerator = torch.sum(torch.square(val_two - val_one))
        denominator = torch.sum(torch.square(val_two))

        return numerator / denominator

    def _create_comparison(
        val_one: TensorOrCollectionType, val_two: TensorOrCollectionType
    ) -> TensorOrCollectionType:
        return recursive_combine(val_one, val_two, _square_head)

    return _create_comparison



"""
Module for defining and managing events in the LLM Compressor.

This module provides an Enum for different event types and a class for creating and
managing events, including methods for calculating event properties and triggering
updates based on specified intervals.
"""

from copy import deepcopy
from dataclasses import dataclass
from enum import Enum, unique
from typing import Optional

from loguru import logger

__all__ = [
    "EventType",
    "Event",
]


@unique
class EventType(Enum):
    """
    An Enum for defining the different types of events that can be triggered
    during model compression lifecycles.
    The purpose of each EventType is to trigger the corresponding
    modifier callback during training or post training pipelines.

    :param INITIALIZE: Event type for initialization.
    :param FINALIZE: Event type for finalization.
    :param BATCH_START: Event type for the start of a batch.
    :param LOSS_CALCULATED: Event type for when loss is calculated.
    :param BATCH_END: Event type for the end of a batch.
    :param OPTIM_PRE_STEP: Event type for pre-optimization step.
    :param OPTIM_POST_STEP: Event type for post-optimization step.
    """

    # training lifecycle
    INITIALIZE = "initialize"
    FINALIZE = "finalize"

    # batch lifecycle
    BATCH_START = "batch_start"
    LOSS_CALCULATED = "loss_calculated"
    BATCH_END = "batch_end"

    # step lifecycle
    OPTIM_PRE_STEP = "optim_pre_step"
    OPTIM_POST_STEP = "optim_post_step"


@dataclass
class Event:
    """
    A class for defining an event that can be triggered during sparsification.

    :param type_: The type of event.
    :type type_: Optional[EventType]
    :param steps_per_epoch: The number of steps per epoch.
    :type steps_per_epoch: Optional[int]
    :param batches_per_step: The number of batches per step where step is an
        optimizer step invocation. For most pathways, these are the same.
        See the invocations_per_step parameter for more details when they are not.
    :type batches_per_step: Optional[int]
    :param invocations_per_step: The number of invocations of the step wrapper
        before optimizer.step was called. Generally can be left as 1 (default).
        For older amp pathways, this is the number of times the scaler wrapper
        was invoked before the wrapped optimizer step function was called to
        handle accumulation in fp16.
    :type invocations_per_step: Optional[int]
    :param global_step: The current global step.
    :type global_step: int
    :param global_batch: The current global batch.
    :type global_batch: int
    """

    type_: Optional[EventType] = None
    steps_per_epoch: Optional[int] = None
    batches_per_step: Optional[int] = None
    invocations_per_step: int = 1
    global_step: int = 0
    global_batch: int = 0

    @property
    def epoch_based(self) -> bool:
        """
        Determines if the event is based on epochs.

        :return: True if the event is based on epochs, False otherwise.
        :rtype: bool
        """
        return self.steps_per_epoch is not None

    @property
    def epoch(self) -> int:
        """
        Calculates the current epoch.

        :raises ValueError: if the event is not epoch based.
        :return: The current epoch.
        :rtype: int
        """
        if not self.epoch_based:
            logger.error("Attempt to access epoch for a non-epoch based event")
            raise ValueError("Event is not epoch based")
        return self.global_step // self.steps_per_epoch

    @property
    def epoch_full(self) -> float:
        """
        Calculates the current epoch with the fraction of the current step.

        :raises ValueError: if the event is not epoch based.
        :return: The current epoch with the fraction of the current step.
        :rtype: float
        """
        if not self.epoch_based:
            logger.error("Attempt to access epoch_full for a non-epoch based event")
            raise ValueError("Event is not epoch based")
        return self.global_step / float(self.steps_per_epoch)

    @property
    def epoch_step(self) -> int:
        """
        Calculates the current step within the current epoch.

        :raises ValueError: if the event is not epoch based.
        :return: The current step within the current epoch.
        :rtype: int
        """
        if not self.epoch_based:
            logger.error("Attempt to access epoch_step for a non-epoch based event")
            raise ValueError("Event is not epoch based")
        return self.global_step % self.steps_per_epoch

    @property
    def epoch_batch(self) -> int:
        """
        Calculates the current batch within the current epoch.

        :raises ValueError: if the event is not epoch based.
        :return: The current batch within the current epoch.
        :rtype: int
        """
        if not self.epoch_based:
            logger.error("Attempt to access epoch_batch for a non-epoch based event")
            raise ValueError("Event is not epoch based")
        batches_per_epoch = (
            self.steps_per_epoch * self.batches_per_step
            if self.batches_per_step
            else self.steps_per_epoch
        )
        return self.global_batch % batches_per_epoch

    @property
    def current_index(self) -> float:
        """
        Calculates the current index of the event.

        :raises ValueError: if the event is not epoch based or
            if the steps per epoch are too many.
        :return: The current index of the event, which is either the global step
            or the epoch with the fraction of the current step.
        :rtype: float
        """
        if not self.epoch_based:
            return self.global_step
        epoch_full = self.epoch_full
        if epoch_full - self.epoch > 1.0:
            logger.error("Too many steps per epoch for epoch based event")
            raise ValueError("Too many steps per epoch for epoch based event")
        return epoch_full

    @current_index.setter
    def current_index(self, value: float):
        """
        Sets the current index of the event.

        :param value: The current index value.
        :type value: float
        """
        logger.debug("Setting current index: {}", value)
        if not self.epoch_based:
            self.global_step = int(value)
            self.global_batch = (
                self.global_step
                if self.batches_per_step is None or self.batches_per_step < 2
                else self.global_step * self.batches_per_step
            )
        else:
            self.global_step = int(value * self.steps_per_epoch)
            self.global_batch = (
                self.global_step
                if self.batches_per_step is None or self.batches_per_step < 2
                else self.global_step * self.batches_per_step
            )

    def should_update(
        self, start: Optional[float], end: Optional[float], update: Optional[float]
    ) -> bool:
        """
        Determines if the event should trigger an update.

        :param start: The start index to check against, set to None to ignore start.
        :type start: Optional[float]
        :param end: The end index to check against, set to None to ignore end.
        :type end: Optional[float]
        :param update: The update interval, set to None or 0.0 to always update,
            otherwise must be greater than 0.0, defaults to None.
        :type update: Optional[float]
        :return: True if the event should trigger an update, False otherwise.
        :rtype: bool
        """
        current = self.current_index
        logger.debug(
            "Checking if event should update: "
            "current_index={}, start={}, end={}, update={}",
            current,
            start,
            end,
            update,
        )
        if start is not None and current < start:
            return False
        if end is not None and current > end:
            return False
        return update is None or update <= 0.0 or current % update < 1e-10

    def new_instance(self, **kwargs) -> "Event":
        """
        Creates a new instance of the event with the provided keyword arguments.

        :param kwargs: Keyword arguments to set in the new instance.
        :return: A new instance of the event with the provided kwargs.
        :rtype: Event
        """
        logger.debug("Creating new instance of event with kwargs: {}", kwargs)
        instance = deepcopy(self)
        for key, value in kwargs.items():
            setattr(instance, key, value)
        return instance



from types import SimpleNamespace

import pytest
import torch

from llmcompressor.utils import (
    ALL_TOKEN,
    DisableQuantization,
    calibration_forward_context,
    convert_to_bool,
    flatten_iterable,
    getattr_chain,
    interpolate,
    patch_attr,
    validate_str_iterable,
)


@pytest.mark.unit
@pytest.mark.parametrize(
    "test_list,output",
    [
        ([], []),
        ([0, 1], [0, 1]),
        ([[0, 1], [2, 3]], [0, 1, 2, 3]),
        ([[0, 1], 2, 3], [0, 1, 2, 3]),
    ],
)
def test_flatten_iterable(test_list, output):
    flattened = flatten_iterable(test_list)
    assert flattened == output


@pytest.mark.unit
@pytest.mark.parametrize(
    "test_bool,output",
    [
        (True, True),
        ("t", True),
        ("T", True),
        ("true)", True),
        ("True", True),
        (1, True),
        ("1", True),
        (False, False),
        ("f", False),
        ("F", False),
        ("false", False),
        ("False", False),
        (0, False),
        ("0", False),
    ],
)
def test_convert_to_bool(test_bool, output):
    converted = convert_to_bool(test_bool)
    assert converted == output


@pytest.mark.unit
@pytest.mark.parametrize(
    "test_list,output",
    [
        (ALL_TOKEN, ALL_TOKEN),
        (ALL_TOKEN.lower(), ALL_TOKEN),
        ([], []),
        ([0, 1], [0, 1]),
        ([[0], [1]], [0, 1]),
    ],
)
def test_validate_str_iterable(test_list, output):
    validated = validate_str_iterable(test_list, "")
    assert validated == output


@pytest.mark.unit
def test_validate_str_iterable_negative():
    with pytest.raises(ValueError):
        validate_str_iterable("will fail", "")


@pytest.mark.unit
@pytest.mark.parametrize(
    "x_cur,x0,x1,y0,y1,inter_func,out",
    [
        (0.0, 0.0, 1.0, 0.0, 5.0, "linear", 0.0),
        (0.0, 0.0, 1.0, 0.0, 5.0, "cubic", 0.0),
        (0.0, 0.0, 1.0, 0.0, 5.0, "inverse_cubic", 0.0),
        (1.0, 0.0, 1.0, 0.0, 5.0, "linear", 5.0),
        (1.0, 0.0, 1.0, 0.0, 5.0, "cubic", 5.0),
        (1.0, 0.0, 1.0, 0.0, 5.0, "inverse_cubic", 5.0),
        (0.5, 0.0, 1.0, 0.0, 5.0, "linear", 2.5),
        (0.5, 0.0, 1.0, 0.0, 5.0, "cubic", 4.375),
        (0.5, 0.0, 1.0, 0.0, 5.0, "inverse_cubic", 1.031),
    ],
)
def test_interpolate(x_cur, x0, x1, y0, y1, inter_func, out):
    interpolated = interpolate(x_cur, x0, x1, y0, y1, inter_func)
    assert abs(out - interpolated) < 0.01


@pytest.mark.unit
def test_getattr_chain():
    base = SimpleNamespace()
    base.a = None
    base.b = SimpleNamespace()
    base.b.c = "value"
    base.b.d = None

    # test base cases
    assert getattr_chain(base, "", None) is None
    with pytest.raises(AttributeError):
        getattr_chain(base, "")

    # test single layer
    assert getattr_chain(base, "a") is None
    assert getattr_chain(base, "a", "default") is None
    assert getattr_chain(base, "b") == base.b

    assert getattr_chain(base, "dne", None) is None
    with pytest.raises(AttributeError):
        getattr_chain(base, "dne")

    # test multi layer
    assert getattr_chain(base, "b.c") == "value"
    assert getattr_chain(base, "b.d") is None
    assert getattr_chain(base, "b.d", "default") is None

    assert getattr_chain(base, "b.d.dne", "default") == "default"
    with pytest.raises(AttributeError):
        getattr_chain(base, "b.d.dne")


@pytest.mark.unit
def test_DisableQuantization():
    model = torch.nn.Linear(1, 1)
    with DisableQuantization(model):
        assert not model.quantization_enabled
    assert model.quantization_enabled


@pytest.mark.unit
def test_calibration_forward_context():
    model = torch.nn.Linear(1, 1)
    model.config = SimpleNamespace()
    model.config.use_cache = True
    model.train()

    with calibration_forward_context(model):
        assert not torch.is_grad_enabled()
        assert not model.config.use_cache
        assert not model.training
    assert torch.is_grad_enabled()
    assert model.config.use_cache
    assert model.training


@pytest.mark.unit
def test_patch_attr():
    # patch, original value
    obj = SimpleNamespace()
    obj.attribute = "original"
    with patch_attr(obj, "attribute", "patched"):
        assert obj.attribute == "patched"
        obj.attribute = "modified"
    assert obj.attribute == "original"

    # patch, no original attribute
    obj = SimpleNamespace()
    with patch_attr(obj, "attribute", "patched"):
        assert obj.attribute == "patched"
        obj.attribute = "modified"
    assert not hasattr(obj, "attribute")



# flake8: noqa

from .dataset_arguments import DatasetArguments
from .model_arguments import ModelArguments
from .recipe_arguments import RecipeArguments
from .training_arguments import TrainingArguments
from .utils import parse_args



import inspect
from typing import Optional

from loguru import logger
from torch.nn import Module
from transformers import AutoModelForCausalLM

__all__ = [
    "SparseAutoModelForCausalLM",
    "get_processor_name_from_model",
]


class SparseAutoModelForCausalLM:
    def from_pretrained(*args, **kwargs):
        logger.warning(
            "SparseAutoModelForCausalLM is deprecated, "
            "please use AutoModelForCausalLM"
        )
        return AutoModelForCausalLM.from_pretrained(*args, **kwargs)


def get_processor_name_from_model(student: Module, teacher: Optional[Module]) -> str:
    """
    Get a processor/tokenizer source used for both student and teacher, assuming
    that they could be shared

    :param student: the student model
    :param teacher: the teacher model
    :return: the source for the processor/tokenizer shared between teacher and model
    """

    if teacher is not None and teacher not in ("disable", "self"):
        student_forward_params = list(
            inspect.signature(student.forward).parameters.keys()
        )
        teacher_forward_params = list(
            inspect.signature(teacher.forward).parameters.keys()
        )
        diff = [p for p in student_forward_params if p not in teacher_forward_params]
        if diff:
            raise RuntimeError(
                "Teacher tokenizer cannot be used for student "
                f"due to missing args: {diff}"
            )
        src_model = teacher
    else:
        src_model = student
    return src_model.config._name_or_path



import inspect
from functools import cached_property
from inspect import _ParameterKind as Kind
from typing import Any, Callable, Dict, List, Union

from compressed_tensors.registry import RegistryMixin
from datasets import Dataset, IterableDataset
from datasets.formatting.formatting import LazyRow
from loguru import logger

from llmcompressor.args import DatasetArguments
from llmcompressor.transformers.finetune.data.data_helpers import (
    LABELS_MASK_VALUE,
    get_custom_datasets_from_path,
    get_raw_dataset,
)
from llmcompressor.transformers.utils.preprocessing_functions import (
    PreprocessingFunctionRegistry,
)
from llmcompressor.typing import DatasetType, Processor
from llmcompressor.utils import import_from_path


class TextGenerationDataset(RegistryMixin):
    """
    Base class for text datasets. Applies the following transformations to a dataset
    in order to prepare the dataset to be loaded by a dataloader

    1. Load dataset from huggingface or local cache
    2. Preprocess dataset according to preprocess function or chat/dataset template
    3. Tokenize dataset using model tokenizer/processor
    4. Apply post processing such as grouping text and/or adding labels for finetuning

    :param dataset_args: configuration settings for dataset loading
    :param split: split from dataset to load, for instance `test` or `train[:5%]`
    :param processor: processor or tokenizer to use on dataset
    """

    # used to mask out the prompt so prompt tokens do not contribute to training loss
    PROMPT_KEY = "prompt"

    def __init__(
        self,
        dataset_args: DatasetArguments,
        split: str,
        processor: Processor,
    ):
        self.dataset_args = dataset_args
        self.split = split
        self.processor = processor

        # get tokenizer
        self.tokenizer = getattr(self.processor, "tokenizer", self.processor)

        if self.tokenizer is not None:
            # fill in pad token
            if not self.tokenizer.pad_token:
                self.tokenizer.pad_token = self.tokenizer.eos_token

            # configure sequence length
            max_seq_length = dataset_args.max_seq_length
            if dataset_args.max_seq_length > self.tokenizer.model_max_length:
                logger.warning(
                    f"The max_seq_length passed ({max_seq_length}) is larger than "
                    f"maximum length for model ({self.tokenizer.model_max_length}). "
                    f"Using max_seq_length={self.tokenizer.model_max_length}."
                )
            self.max_seq_length = min(
                dataset_args.max_seq_length, self.tokenizer.model_max_length
            )

            # configure padding
            self.padding = (
                False
                if self.dataset_args.concatenate_data
                else "max_length"
                if self.dataset_args.pad_to_max_length
                else False
            )

        else:
            self.max_seq_length = None
            self.padding = False

    def __call__(self, add_labels: bool = True) -> DatasetType:
        dataset = self.dataset_args.dataset

        if isinstance(dataset, str):
            # load dataset: load from huggingface or disk
            dataset = self.load_dataset()
        logger.debug(f"Raw dataset: {get_columns(dataset)}")

        if self.preprocess is not None:
            # preprocess: apply template or preprocessing function
            dataset = self.map(
                dataset,
                self.preprocess,
                batched=False,
                num_proc=self.dataset_args.preprocessing_num_workers,
                load_from_cache_file=not self.dataset_args.overwrite_cache,
                desc="Preprocessing",
            )
            logger.debug(f"Dataset after preprocessing: {get_columns(dataset)}")

        # rename and remove columns match processor kwargs
        dataset = self.rename_columns(dataset)
        logger.debug(f"Dataset after column renaming: {get_columns(dataset)}")

        # use processor.model_input_names to determine if the ds is already tokenized
        model_input_names = getattr(self.processor, "model_input_names", ["input_ids"])
        if not any(col_name in model_input_names for col_name in get_columns(dataset)):
            # tokenize/ process
            dataset = self.filter_tokenizer_args(dataset)
            logger.debug(f"Tokenizer args after filtering: {get_columns(dataset)}")
            dataset = self.map(
                dataset,
                self.tokenize,
                batched=False,  # batching is not well supported for vision processors
                keep_in_memory=True,  # bug occurs when not batched and not in memory,
                # subsequent ds.map calls are always batched,
                # regardless of `batched` argument
                remove_columns=get_columns(dataset),  # assumes that input names
                # and output names are disjoint
                num_proc=self.dataset_args.preprocessing_num_workers,
                load_from_cache_file=not self.dataset_args.overwrite_cache,
                desc="Tokenizing",
            )
            logger.debug(f"Model kwargs after tokenizing: {get_columns(dataset)}")

        if self.dataset_args.concatenate_data:
            # postprocess: group text
            dataset = self.map(
                dataset,
                self.group_text,
                batched=True,
                num_proc=self.dataset_args.preprocessing_num_workers,
                load_from_cache_file=not self.dataset_args.overwrite_cache,
                desc="Concatenating data",
            )
            logger.debug(f"Model kwargs after concatenating: {get_columns(dataset)}")

        if add_labels:
            # postprocess: add labels
            dataset = self.map(
                dataset,
                self.add_labels,
                batched=False,  # not compatible with batching, need row lengths
                num_proc=self.dataset_args.preprocessing_num_workers,
                load_from_cache_file=not self.dataset_args.overwrite_cache,
                desc="Adding labels",
            )
            logger.debug(f"Model kwargs after adding labels: {get_columns(dataset)}")

        elif self.PROMPT_KEY in get_columns(dataset):
            dataset = dataset.remove_columns(self.PROMPT_KEY)
            logger.debug("Removed prompt key")

        logger.debug(f"Model kwargs after postprocessing: {get_columns(dataset)}")
        return dataset

    def load_dataset(self):
        """
        Load the raw dataset from Hugging Face, using cached copy if available

        :param cache_dir: disk location to search for cached dataset
        :return: the requested dataset
        """
        if self.dataset_args.dataset_path is not None:
            if self.dataset_args.dvc_data_repository is not None:
                self.dataset_args.raw_kwargs["storage_options"] = {
                    "url": self.dataset_args.dvc_data_repository
                }
                self.dataset_args.raw_kwargs["data_files"] = (
                    self.dataset_args.dataset_path
                )
            else:
                self.dataset_args.raw_kwargs["data_files"] = (
                    get_custom_datasets_from_path(
                        self.dataset_args.dataset_path,
                        self.dataset_args.dataset
                        if hasattr(self.dataset_args, "dataset")
                        else self.dataset_args.dataset_name,
                    )
                )

        logger.debug(f"Loading dataset {self.dataset_args.dataset}")
        return get_raw_dataset(
            self.dataset_args,
            None,
            split=self.split,
            streaming=self.dataset_args.streaming,
            **self.dataset_args.raw_kwargs,
        )

    @cached_property
    def preprocess(self) -> Union[Callable[[LazyRow], Any], None]:
        """
        The function must return keys which correspond to processor/tokenizer kwargs,
        optionally including PROMPT_KEY
        """
        preprocessing_func = self.dataset_args.preprocessing_func

        if callable(preprocessing_func):
            return preprocessing_func

        if isinstance(preprocessing_func, str):
            if ":" in preprocessing_func:
                # load func_name from "/path/to/file.py:func_name"
                return import_from_path(preprocessing_func)
            else:
                # load from the registry
                return PreprocessingFunctionRegistry.get_value_from_registry(
                    name=preprocessing_func
                )

        return self.dataset_template

    @property
    def dataset_template(self) -> Union[Callable[[Any], Any], None]:
        return None

    def rename_columns(self, dataset: DatasetType) -> DatasetType:
        # rename columns to match processor/tokenizer kwargs
        column_names = get_columns(dataset)
        if self.dataset_args.text_column in column_names and "text" not in column_names:
            logger.debug(f"Renaming column `{self.dataset_args.text_column}` to `text`")
            dataset = dataset.rename_column(self.dataset_args.text_column, "text")

        return dataset

    def filter_tokenizer_args(self, dataset: DatasetType) -> DatasetType:
        # assumes that inputs are not passed via self.processor.__call__ args and kwargs
        signature = inspect.signature(self.processor.__call__)
        tokenizer_args = set(
            key
            for key, param in signature.parameters.items()
            if param.kind not in (Kind.VAR_POSITIONAL, Kind.VAR_KEYWORD)
        )
        logger.debug(
            f"Found processor args `{tokenizer_args}`. Removing all other columns"
        )

        column_names = get_columns(dataset)
        return dataset.remove_columns(
            list(set(column_names) - set(tokenizer_args) - set([self.PROMPT_KEY]))
        )

    def tokenize(self, data: LazyRow) -> Dict[str, Any]:
        # separate prompt
        prompt = data.pop(self.PROMPT_KEY, None)

        # tokenize
        data = self.processor(
            **data,
            padding=self.padding,
            max_length=self.max_seq_length,
            truncation=True,
        )

        # store unpadded prompt so we can mask out correct number of elements in labels
        if prompt is not None:
            data[self.PROMPT_KEY] = self.processor(
                text=prompt,
                max_length=self.max_seq_length,
                truncation=True,
            )["input_ids"]

        return data

    def group_text(self, data: LazyRow) -> Dict[str, Any]:
        concatenated_data = {k: sum(data[k], []) for k in data.keys()}
        total_length = len(concatenated_data[list(data.keys())[0]])
        total_length = (total_length // self.max_seq_length) * self.max_seq_length
        result = {
            k: [
                t[i : i + self.max_seq_length]
                for i in range(0, total_length, self.max_seq_length)
            ]
            for k, t in concatenated_data.items()
        }
        return result

    def add_labels(self, data: LazyRow) -> LazyRow:
        if "pixel_values" in data:
            raise NotImplementedError(
                "Label masking for vision datasets has not been implemented yet"
            )

        # if the dataset uses prompts, mask them out so they don't contribute
        # to the loss calculation
        prompt_len = 0
        if self.PROMPT_KEY in data:
            prompt_len = len(data[self.PROMPT_KEY])
        data["labels"] = data["input_ids"].copy()
        data["labels"][:prompt_len] = [LABELS_MASK_VALUE] * prompt_len

        # mask out padding in the labels as well
        padding = len(data["attention_mask"]) - sum(data["attention_mask"])
        if padding > 0:
            data["labels"][-padding:] = [LABELS_MASK_VALUE] * padding
        return data

    def map(
        self,
        dataset: Union[Dataset, IterableDataset],
        function: Callable[[Any], Any],
        **kwargs,
    ) -> Union[Dataset, IterableDataset]:
        """
        Wrapper function around Dataset.map and IterableDataset.map.

        If the dataset is streaming (in the case of IterableDataset), non-applicable
        arguments are ignored and the dataset features are resolved
        """
        if isinstance(dataset, IterableDataset):
            # remove arguments that don't apply to streaming
            kwargs.pop("num_proc", None)
            kwargs.pop("load_from_cache_file", None)
            kwargs.pop("desc", None)
            kwargs.pop("keep_in_memory", None)

        dataset = dataset.map(function, **kwargs)

        if isinstance(dataset, IterableDataset):
            dataset = dataset._resolve_features()

        return dataset


def get_columns(dataset: DatasetType) -> List[str]:
    column_names = dataset.column_names
    if isinstance(column_names, dict):
        column_names = sum(column_names.values(), [])

    return column_names



import os
import shutil
import tempfile
from typing import List

import pytest

try:
    import wandb
except Exception:
    wandb = None


os.environ["NM_TEST_MODE"] = "True"
os.environ["NM_TEST_LOG_DIR"] = "nm_temp_test_logs"


def _get_files(directory: str) -> List[str]:
    list_filepaths = []
    for root, dirs, files in os.walk(directory):
        for file in files:
            list_filepaths.append(os.path.join(os.path.abspath(root), file))
    return list_filepaths


def _files_size_mb(path_list: List[str]) -> int:
    files_size = 0
    for file_path in path_list:
        try:
            files_size += os.path.getsize(file_path)
        # if file is deleted between the time we get the list of files
        # and the time we get the size of the file, ignore it
        except FileNotFoundError:
            pass

    return files_size / 1024 / 1024


@pytest.fixture(scope="session", autouse=True)
def check_for_created_files():
    start_files_root = _get_files(directory=r".")
    start_files_temp = _get_files(directory=tempfile.gettempdir())
    yield
    if wandb:
        wandb.finish()
    log_dir = os.environ.get("NM_TEST_LOG_DIR")
    if os.path.isdir(log_dir):
        shutil.rmtree(log_dir)

    # allow creation of __pycache__ directories
    end_files_root = [
        f_path for f_path in _get_files(directory=r".") if "__pycache__" not in f_path
    ]
    end_files_temp = _get_files(directory=tempfile.gettempdir())
    # assert no files created in root directory while running
    # the pytest suite
    assert len(start_files_root) >= len(end_files_root), (
        f"{len(end_files_root) - len(start_files_root)} "
        f"files created in current working "
        f"directory during pytest run. "
        f"Created files: {set(end_files_root) - set(start_files_root)}"
    )

    max_allowed_sized_temp_files_megabytes = 1
    end_files_temp = _get_files(directory=tempfile.gettempdir())
    created_temp_files = set(end_files_temp) - set(start_files_temp)
    # pytest temp files are automatically deleted, exclude from size calculation
    created_temp_files = [
        f_path for f_path in created_temp_files if "pytest-of" not in f_path
    ]

    # assert no more than 1 megabyte of temp files created in temp directory
    # while running the pytest suite (excluding files created by pytest)
    size_of_temp_files_megabytes = _files_size_mb(created_temp_files)
    assert max_allowed_sized_temp_files_megabytes >= size_of_temp_files_megabytes, (
        f"{size_of_temp_files_megabytes} "
        f"megabytes of temp files created in temp directory during pytest run. "
        f"Created files: {set(end_files_temp) - set(start_files_temp)}"
    )


@pytest.fixture(autouse=True, scope="function")
def setup_fresh_session():
    """
    setup any state tied to the execution of the given method in a
    class.  setup_method is invoked for every test method of a class.
    """
    from llmcompressor.core import reset_session

    # start with a clean session for each test
    reset_session()
    yield
    # reset the session after each test
    reset_session()



import tempfile

import pytest
import yaml

from llmcompressor.modifiers.obcq.base import SparseGPTModifier
from llmcompressor.recipe import Recipe
from tests.llmcompressor.helpers import valid_recipe_strings


@pytest.mark.parametrize("recipe_str", valid_recipe_strings())
def test_recipe_create_instance_accepts_valid_recipe_string(recipe_str):
    recipe = Recipe.create_instance(recipe_str)
    assert recipe is not None, "Recipe could not be created from string"


@pytest.mark.parametrize("recipe_str", valid_recipe_strings())
def test_recipe_create_instance_accepts_valid_recipe_file(recipe_str):
    content = yaml.safe_load(recipe_str)
    with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml") as f:
        yaml.dump(content, f)
        recipe = Recipe.create_instance(f.name)
        assert recipe is not None, "Recipe could not be created from file"


@pytest.mark.parametrize("recipe_str", valid_recipe_strings())
def test_serialization(recipe_str):
    recipe_instance = Recipe.create_instance(recipe_str)
    serialized_recipe = recipe_instance.yaml()
    recipe_from_serialized = Recipe.create_instance(serialized_recipe)

    expected_dict = recipe_instance.dict()
    actual_dict = recipe_from_serialized.dict()

    assert expected_dict == actual_dict


def test_recipe_creates_correct_modifier():
    start = 1
    end = 10
    targets = "__ALL_PRUNABLE__"

    yaml_str = f"""
        test_stage:
            pruning_modifiers:
                ConstantPruningModifier:
                    start: {start}
                    end: {end}
                    targets: {targets}
        """

    recipe_instance = Recipe.create_instance(yaml_str)

    stage_modifiers = recipe_instance.create_modifier()
    assert len(stage_modifiers) == 1
    assert len(modifiers := stage_modifiers[0].modifiers) == 1
    from llmcompressor.modifiers.pruning.constant import ConstantPruningModifier

    assert isinstance(modifier := modifiers[0], ConstantPruningModifier)
    assert modifier.start == start
    assert modifier.end == end


def test_recipe_can_be_created_from_modifier_instances():
    modifier = SparseGPTModifier(
        sparsity=0.5,
    )
    group_name = "dummy"

    # for pep8 compliance
    recipe_str = (
        f"{group_name}_stage:\n"
        "   pruning_modifiers:\n"
        "       SparseGPTModifier:\n"
        "           sparsity: 0.5\n"
    )

    expected_recipe_instance = Recipe.create_instance(recipe_str)
    expected_modifiers = expected_recipe_instance.create_modifier()

    actual_recipe_instance = Recipe.create_instance(
        [modifier], modifier_group_name=group_name
    )
    actual_modifiers = actual_recipe_instance.create_modifier()

    # assert num stages is the same
    assert len(actual_modifiers) == len(expected_modifiers)

    # assert num modifiers in each stage is the same
    assert len(actual_modifiers[0].modifiers) == len(expected_modifiers[0].modifiers)

    # assert modifiers in each stage are the same type
    # and have the same parameters
    for actual_modifier, expected_modifier in zip(
        actual_modifiers[0].modifiers, expected_modifiers[0].modifiers
    ):
        assert isinstance(actual_modifier, type(expected_modifier))
        assert actual_modifier.model_dump() == expected_modifier.model_dump()



import argparse

from datasets import load_dataset
from transformers import AutoModelForCausalLM, AutoTokenizer

from llmcompressor import oneshot
from llmcompressor.modifiers.obcq import SparseGPTModifier
from llmcompressor.modifiers.quantization import QuantizationModifier

# Configuration
MODEL_ID = "meta-llama/Meta-Llama-3-8B-Instruct"
DATASET_ID = "HuggingFaceH4/ultrachat_200k"
DATASET_SPLIT = "train_sft"
NUM_CALIBRATION_SAMPLES = 512
MAX_SEQUENCE_LENGTH = 2048


def parse_args():
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(description="Apply compression to a model")
    parser.add_argument("--fp8", action="store_true", help="Enable FP8 compression")
    return parser.parse_args()


def preprocess(example):
    """Preprocess dataset examples."""
    return {"text": tokenizer.apply_chat_template(example["messages"], tokenize=False)}


def tokenize(sample):
    """Tokenize dataset examples."""
    return tokenizer(
        sample["text"],
        padding=False,
        max_length=MAX_SEQUENCE_LENGTH,
        truncation=True,
        add_special_tokens=False,
    )


def get_recipe(fp8_enabled):
    """Generate the compression recipe and save directory based on the FP8 flag."""
    base_recipe = [
        SparseGPTModifier(
            sparsity=0.5,
            mask_structure="2:4",
            sequential_update=True,
            targets=[r"re:model.layers.\d*$"],
        )
    ]
    save_dir = MODEL_ID.split("/")[1] + "2of4-sparse"

    if fp8_enabled:
        base_recipe.append(
            QuantizationModifier(
                targets=["Linear"],
                ignore=["lm_head"],
                scheme="FP8_DYNAMIC",
            )
        )
        save_dir = MODEL_ID.split("/")[1] + "2of4-W8A8-FP8-Dynamic-Per-Token"

        # check that asymmetric quantization is not being used
        q_scheme = base_recipe[1].scheme
        if not isinstance(q_scheme, str) and not q_scheme["weights"].symmetric:
            raise ValueError(
                "Asymmetric quantization with 2of4 sparsity is not supported by vLLM. "
                "Please use symmetric quantization"
            )

    return base_recipe, save_dir


# Parse arguments
args = parse_args()

# Load model and tokenizer
model = AutoModelForCausalLM.from_pretrained(
    MODEL_ID, device_map="auto", torch_dtype="auto"
)
tokenizer = AutoTokenizer.from_pretrained(MODEL_ID)

# Load and preprocess dataset
ds = (
    load_dataset(DATASET_ID, split=DATASET_SPLIT)
    .shuffle(seed=42)
    .select(range(NUM_CALIBRATION_SAMPLES))
)
ds = ds.map(preprocess)
ds = ds.map(tokenize, remove_columns=ds.column_names)

# Get compression recipe and save directory
recipe, save_dir = get_recipe(args.fp8)

# Apply compression
oneshot(
    model=model,
    dataset=ds,
    recipe=recipe,
    max_seq_length=MAX_SEQUENCE_LENGTH,
    num_calibration_samples=NUM_CALIBRATION_SAMPLES,
)

# Validate the compressed model
print("\n========== SAMPLE GENERATION ==============")
input_ids = tokenizer("Hello my name is", return_tensors="pt").input_ids.to("cuda")
output = model.generate(input_ids, max_new_tokens=100)
print(tokenizer.decode(output[0]))
print("==========================================\n")

# Save compressed model and tokenizer
model.save_pretrained(save_dir)
tokenizer.save_pretrained(save_dir)



"""
Helper functions for retrieving information related to model sparsification
"""

import json
from typing import Dict, Optional

import torch
from accelerate.accelerator import get_state_dict_offloaded_model
from torch.nn import Module
from tqdm import tqdm

from llmcompressor.pytorch.utils.helpers import get_quantized_layers, tensor_sparsity

__all__ = [
    "ModuleSparsificationInfo",
]


class ModuleSparsificationInfo:
    """
    Helper class for providing information related to torch Module parameters
    and the amount of sparsification applied. Includes information for pruning
    and quantization

    :param module: torch Module to analyze
    :param state_dict: optional state_dict to analyze in place of the torch model. This
    is used when analyzing an FSDP model, where the full weights may not be accessible
    """

    def __init__(
        self, module: Module, state_dict: Optional[Dict[str, torch.Tensor]] = None
    ):
        self.module = module

        if state_dict is not None:
            # when analyzing an FSDP model, the state_dict does not differentiate
            # between trainable and non-trainable parameters
            # (e.g. it can contain buffers) this means that the
            # self.trainable_parameters may be overestimated
            self.trainable_params = state_dict
        else:
            if hasattr(module, "_hf_hook"):
                self.trainable_params = get_state_dict_offloaded_model(module)
            else:
                self.trainable_params = {
                    k: v for k, v in self.module.named_parameters() if v.requires_grad
                }

    def __str__(self):
        return json.dumps(
            {
                "params_summary": {
                    "total": self.params_total,
                    "sparse": self.params_sparse,
                    "sparsity_percent": self.params_sparse_percent,
                    "quantized": self.params_quantized,
                    "quantized_percent": self.params_quantized_percent,
                },
                "params_info": self.params_info,
            }
        )

    @property
    def params_total(self) -> int:
        """
        :return: total number of trainable parameters in the model
        """
        return sum(torch.numel(param) for param in self.trainable_params.values())

    @property
    def params_sparse(self) -> int:
        """
        :return: total number of sparse (0) trainable parameters in the model
        """
        return sum(
            round(tensor_sparsity(param).item() * torch.numel(param))
            for param in tqdm(
                self.trainable_params.values(), desc="Calculating model sparsity"
            )
        )

    @property
    def params_sparse_percent(self) -> float:
        """
        :return: percent of sparsified parameters in the entire model
        """
        return self.params_sparse / float(self.params_total) * 100

    @property
    def params_quantized(self) -> int:
        """
        :return: number of parameters across quantized layers
        """
        num_params = 0
        for name, layer in get_quantized_layers(self.module):
            if getattr(layer, "weight", None) is not None:
                num_params += torch.numel(layer.weight)
            if getattr(layer, "bias", None) is not None:
                num_params += torch.numel(layer.bias)

        return num_params

    @property
    def params_quantized_percent(self) -> float:
        """
        :return: percentage of parameters that have been quantized
        """
        return self.params_quantized / float(self.params_total) * 100



import unittest

import pytest

from llmcompressor.modifiers.factory import ModifierFactory
from llmcompressor.modifiers.logarithmic_equalization.base import (
    LogarithmicEqualizationModifier,
)
from llmcompressor.modifiers.smoothquant.base import SmoothQuantModifier
from tests.llmcompressor.modifiers.conf import setup_modifier_factory


@pytest.mark.unit
class TestLogarithmicEqualizationIsRegistered(unittest.TestCase):
    def setUp(self):
        self.kwargs = dict(
            smoothing_strength=0.3,
            mappings=[(["layer1", "layer2"], "layer3")],
        )
        setup_modifier_factory()

    def test_log_equalization_is_registered(self):
        modifier = ModifierFactory.create(
            type_="LogarithmicEqualizationModifier",
            allow_experimental=False,
            allow_registered=True,
            **self.kwargs,
        )

        self.assertIsInstance(
            modifier,
            LogarithmicEqualizationModifier,
            "PyTorch LogarithmicEqualizationModifier not registered",
        )

        self.assertIsInstance(modifier, SmoothQuantModifier)
        self.assertEqual(modifier.smoothing_strength, self.kwargs["smoothing_strength"])
        self.assertEqual(modifier.mappings, self.kwargs["mappings"])



import unittest

import pytest
import torch
from datasets import IterableDataset, load_dataset
from parameterized import parameterized

from llmcompressor.args import DatasetArguments
from llmcompressor.datasets import format_calibration_data, get_processed_dataset
from llmcompressor.transformers import TextGenerationDataset


@pytest.mark.unit
class TestConcentrationTokenization(unittest.TestCase):
    def setUp(self):
        self.dataset_args = DatasetArguments(
            dataset="wikitext",
            dataset_config_name="wikitext-2-raw-v1",
            concatenate_data=True,
        )

    @pytest.fixture(autouse=True)
    def prepare_fixture(self, tiny_llama_tokenizer):
        self.tiny_llama_tokenizer = tiny_llama_tokenizer

    def test_concatenation_tokenization(self):
        wiki_manager = TextGenerationDataset.load_from_registry(
            self.dataset_args.dataset,
            dataset_args=self.dataset_args,
            split="train[:5%]",
            processor=self.tiny_llama_tokenizer,
        )
        raw_dataset = wiki_manager.load_dataset()
        self.assertGreater(len(raw_dataset), 0)
        self.assertEqual(raw_dataset.split, "train[:5%]")
        self.assertEqual(raw_dataset.info.config_name, "wikitext-2-raw-v1")
        tokenized_dataset = wiki_manager()
        self.assertIn("input_ids", tokenized_dataset.features)
        self.assertIn("labels", tokenized_dataset.features)
        for i in range(len(tokenized_dataset)):
            self.assertEqual(
                len(tokenized_dataset[i]["input_ids"]), wiki_manager.max_seq_length
            )


@pytest.mark.unit
class TestNoPaddingTokenization(unittest.TestCase):
    def setUp(self):
        self.dataset_args = DatasetArguments(
            dataset="open_platypus", pad_to_max_length=False
        )

    @pytest.fixture(autouse=True)
    def prepare_fixture(self, tiny_llama_tokenizer):
        self.tiny_llama_tokenizer = tiny_llama_tokenizer

    @pytest.mark.usefixtures("tiny_llama_tokenizer")
    def test_no_padding_tokenization(self):
        op_manager = TextGenerationDataset.load_from_registry(
            self.dataset_args.dataset,
            dataset_args=self.dataset_args,
            split="train[5%:7%]",
            processor=self.tiny_llama_tokenizer,
        )
        dataset = op_manager.load_dataset()  # load
        dataset = op_manager.map(  # preprocess
            dataset,
            op_manager.preprocess,
            batched=False,
            num_proc=op_manager.dataset_args.preprocessing_num_workers,
        )
        dataset = op_manager.rename_columns(dataset)  # rename
        self.assertGreater(len(dataset), 0)
        ex_item = dataset[0]["text"]
        self.assertIn("Below is an instruction that describes a task", ex_item)

        self.assertEqual(dataset.split, "train[5%:7%]")
        tokenized_dataset = op_manager()
        self.assertIn("input_ids", tokenized_dataset.features)
        self.assertIn("labels", tokenized_dataset.features)
        print(tokenized_dataset[0]["input_ids"])

        for i in range(len(tokenized_dataset)):
            self.assertLessEqual(
                len(tokenized_dataset[i]["input_ids"]), op_manager.max_seq_length
            )


@pytest.mark.unit
class TestMaxSeqLenClipped(unittest.TestCase):
    def setUp(self):
        self.dataset_args = DatasetArguments(
            dataset="open_platypus", max_seq_length=4096
        )

    @pytest.fixture(autouse=True)
    def prepare_fixture(self, tiny_llama_tokenizer):
        self.tiny_llama_tokenizer = tiny_llama_tokenizer

    def test_max_seq_len_clipped(self):
        op_manager = TextGenerationDataset.load_from_registry(
            self.dataset_args.dataset,
            dataset_args=self.dataset_args,
            split="train[95%:]",
            processor=self.tiny_llama_tokenizer,
        )

        self.assertEqual(
            op_manager.max_seq_length, self.tiny_llama_tokenizer.model_max_length
        )


@pytest.mark.unit
class TestDatasetKwargsAndPercent(unittest.TestCase):
    def setUp(self):
        self.dataset_args = DatasetArguments(
            dataset="wikitext",
            raw_kwargs={
                "data_files": {
                    "train": "wikitext-2-raw-v1/train-00000-of-00001.parquet"
                }
            },
        )

    @pytest.fixture(autouse=True)
    def prepare_fixture(self, tiny_llama_tokenizer):
        self.tiny_llama_tokenizer = tiny_llama_tokenizer

    def test_dataset_kwargs_and_percentages(self):
        c4_manager_a = TextGenerationDataset.load_from_registry(
            self.dataset_args.dataset,
            dataset_args=self.dataset_args,
            split="train[5%:6%]",
            processor=self.tiny_llama_tokenizer,
        )
        raw_dataset_a = c4_manager_a.load_dataset()

        c4_manager_b = TextGenerationDataset.load_from_registry(
            self.dataset_args.dataset,
            dataset_args=self.dataset_args,
            split="train[6%:8%]",
            processor=self.tiny_llama_tokenizer,
        )
        raw_dataset_b = c4_manager_b.load_dataset()

        self.assertEqual(len(raw_dataset_b), 2 * len(raw_dataset_a))


@pytest.mark.unit
class TestDatasets(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def prepare_fixture(self, tiny_llama_tokenizer):
        self.tiny_llama_tokenizer = tiny_llama_tokenizer

    @parameterized.expand(
        [
            ["ptb", "penn_treebank", "train[:5%]", False],
            ["gsm8k", "main", "train[:5%]", True],
            ["ultrachat_200k", "default", "train_sft[:1%]", False],
        ]
    )
    def test_datasets(self, dataset_key, dataset_config, split, do_concat):
        dataset_args = DatasetArguments(
            dataset=dataset_key,
            dataset_config_name=dataset_config,
            concatenate_data=do_concat,
            trust_remote_code_data=True,
        )
        manager = TextGenerationDataset.load_from_registry(
            dataset_args.dataset,
            dataset_args=dataset_args,
            split=split,
            processor=self.tiny_llama_tokenizer,
        )
        raw_dataset = manager.load_dataset()
        self.assertGreater(len(raw_dataset), 0)
        self.assertEqual(raw_dataset.split, split)
        self.assertEqual(raw_dataset.info.config_name, dataset_config)

        tokenized_dataset = manager()
        self.assertIn("input_ids", tokenized_dataset.features)
        self.assertIn("labels", tokenized_dataset.features)
        for i in range(len(tokenized_dataset)):
            if do_concat:
                self.assertEqual(
                    len(tokenized_dataset[i]["input_ids"]), manager.max_seq_length
                )
            else:
                self.assertLessEqual(
                    len(tokenized_dataset[i]["input_ids"]), manager.max_seq_length
                )


@pytest.mark.skip("Dataset load broken on Hugging Face")
@pytest.mark.unit
class TestEvol(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def prepare_fixture(self, tiny_llama_tokenizer):
        self.tiny_llama_tokenizer = tiny_llama_tokenizer

    def setUp(self):
        self.dataset_args = DatasetArguments(
            dataset="evolcodealpaca",
            dataset_config_name=None,
            concatenate_data=False,
        )

    def test_evol(self):
        evol_manager = TextGenerationDataset.load_from_registry(
            self.dataset_args.dataset,
            dataset_args=self.dataset_args,
            split="train[:2%]",
            processor=self.tiny_llama_tokenizer,
        )
        raw_dataset = evol_manager.load_dataset()
        self.assertGreater(len(raw_dataset), 0)
        self.assertEqual(raw_dataset.split, "train[:2%]")

        tokenized_dataset = evol_manager()
        self.assertIn("input_ids", tokenized_dataset.features)
        self.assertIn("labels", tokenized_dataset.features)
        for i in range(len(tokenized_dataset)):
            self.assertLessEqual(
                len(tokenized_dataset[i]["input_ids"]), evol_manager.max_seq_length
            )


@pytest.mark.unit
class TestStreamLoading(unittest.TestCase):
    def setUp(self):
        self.dataset_args = DatasetArguments(
            dataset="wikitext",
            dataset_config_name="wikitext-2-raw-v1",
            concatenate_data=True,
            streaming=True,
        )

    @pytest.fixture(autouse=True)
    def prepare_fixture(self, tiny_llama_tokenizer):
        self.tiny_llama_tokenizer = tiny_llama_tokenizer

    def test_stream_loading(self):
        manager = TextGenerationDataset.load_from_registry(
            self.dataset_args.dataset,
            dataset_args=self.dataset_args,
            split="train",
            processor=self.tiny_llama_tokenizer,
        )

        processed = manager()
        self.assertIsInstance(processed, IterableDataset)
        with pytest.raises(TypeError):
            # in streaming mode we don't know the length of the dataset
            _ = len(processed)

        # confirm tokenization of streamed item works correctly
        item = next(iter(processed))
        self.assertIn("labels", item)
        self.assertEqual(len(item["input_ids"]), manager.max_seq_length)


@pytest.mark.unit
class TestSplitLoading(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def prepare_fixture(self, tiny_llama_tokenizer):
        self.tiny_llama_tokenizer = tiny_llama_tokenizer

    @parameterized.expand([["train[95%:]"], [{"train": "train[:5%]"}]])
    def test_split_loading(self, split_def):
        dataset_args = DatasetArguments(
            dataset="open_platypus",
            splits=split_def,
            trust_remote_code_data=True,
        )

        dataset = get_processed_dataset(
            dataset_args=dataset_args, processor=self.tiny_llama_tokenizer
        )

        assert dataset is not None
        self.assertIsInstance(dataset, dict)


@pytest.mark.unit
class TestTokenizationDataset(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def prepare_fixture(self, tiny_llama_tokenizer):
        self.tiny_llama_tokenizer = tiny_llama_tokenizer
        dataset = load_dataset("garage-bAInd/Open-Platypus")["train"]
        self.num_calib_samples = 64
        self.max_seq_len = 512
        self.dataset = dataset.shuffle(seed=42).select(range(self.num_calib_samples))

    def test_load_tokenized_data(self):
        def preprocess(sample):
            concat_text = "INPUT: " + sample.get("input", "")
            concat_text += "INSTRUCTIONS: " + sample.get("instruction", "")
            concat_text += "OUTPUT: " + sample.get("output", "")

            return self.tiny_llama_tokenizer(
                concat_text, padding=False, max_length=self.max_seq_len, truncation=True
            )

        tokenized_dataset = self.dataset.map(
            preprocess, remove_columns=["input", "output", "instruction", "data_source"]
        )

        dataset_args = DatasetArguments(
            dataset=tokenized_dataset, shuffle_calibration_samples=False
        )

        dataset = get_processed_dataset(
            dataset_args=dataset_args,
            processor=self.tiny_llama_tokenizer,
            do_oneshot=True,
            do_train=False,
        )
        calib_dataset = dataset["calibration"]

        self.assertEqual(len(calib_dataset), self.num_calib_samples)
        data_cols = calib_dataset.column_names
        self.assertEqual(len(data_cols), 2)
        self.assertIn("input_ids", data_cols)
        self.assertIn("attention_mask", data_cols)

        # confirm turning shuffle off works

        calib_dataloader = format_calibration_data(
            tokenized_dataset=calib_dataset,
            num_calibration_samples=self.num_calib_samples,
            do_shuffle=dataset_args.shuffle_calibration_samples,
        )
        self.assertEqual(len(calib_dataloader), self.num_calib_samples)
        dataloader_sample = next(iter(calib_dataloader))["input_ids"]
        diff = dataloader_sample - torch.Tensor(calib_dataset[0]["input_ids"])
        self.assertEqual(torch.sum(diff), 0)



"""
A library for compressing large language models utilizing the latest techniques and
research in the field for both training aware and post training techniques.

The library is designed to be flexible and easy to use on top of
PyTorch and HuggingFace Transformers, allowing for quick experimentation.
"""

# flake8: noqa

from .logger import LoggerConfig, configure_logger, logger
from .version import __version__, version

__all__ = [
    "__version__",
    "version",
    "configure_logger",
    "logger",
    "LoggerConfig",
]

from llmcompressor.core.session_functions import (
    active_session,
    callbacks,
    create_session,
    reset_session,
)
from llmcompressor.entrypoints import Oneshot, oneshot, train



from contextlib import nullcontext as does_not_raise

import pytest

from llmcompressor.metrics.utils import FrequencyManager


@pytest.fixture(scope="function")
def frequency_manager():
    return FrequencyManager(log_frequency=1)


@pytest.mark.parametrize(
    "kwargs, expectation",
    [
        (
            {"log_frequency": 0},
            pytest.raises(ValueError, match="must be greater than 0, given 0"),
        ),
        (
            {"log_frequency": -1},
            pytest.raises(ValueError, match="must be greater than 0, given -1"),
        ),
        (
            {"log_frequency": True},
            pytest.raises(TypeError, match="must be a number or None"),
        ),
        (
            {"log_frequency": []},
            pytest.raises(TypeError, match="must be a number or None"),
        ),
        (
            {"log_frequency": {}},
            pytest.raises(TypeError, match="must be a number or None"),
        ),
        ({"log_frequency": 1}, does_not_raise()),
        ({"log_frequency": None}, does_not_raise()),
    ],
)
def test_frequency_manager_creation(kwargs, expectation):
    with expectation:
        _ = FrequencyManager(**kwargs)


@pytest.mark.parametrize(
    "step, expectation",
    [
        (0.1, does_not_raise()),
        (0, does_not_raise()),
        (-1, pytest.raises(ValueError, match="must be greater than or equal to 0")),
        (True, pytest.raises(TypeError, match="must be a number or None")),
        ([], pytest.raises(TypeError, match="must be a number or None")),
        ({}, pytest.raises(TypeError, match="must be a number or None")),
    ],
)
class TestFrequencyManagerUpdationUtilities:
    def test_model_updated(self, frequency_manager, step, expectation):
        # test that model_updated sets last_model_update_step
        # to the given step

        with expectation:
            frequency_manager.model_updated(step=step)
            assert frequency_manager.last_model_update_step == step

    def test_log_written(self, frequency_manager, step, expectation):
        # test that log_written sets last_log_step
        # to the given step

        with expectation:
            frequency_manager.log_written(step=step)
            assert frequency_manager.last_log_step == step


def _log_ready_test_cases():
    # test cases for log_ready

    # each test case is a tuple of:
    #   (log_frequency, current_log_step, last_log_step,
    #       last_model_update_step, check_model_update, expected)

    return [
        # None values should give True
        (0.1, None, None, None, False, True),
        (0.1, None, 1, 1, False, True),
        (0.1, 1, None, 1, False, True),
        (0.1, 0.3, 0.2, None, False, True),
        (0.1, 0.3, 0.2, None, True, True),
        # log frequency is None
        (None, 1, 2, 3, False, False),
        (None, 1, 2, 3, True, False),
        # cadence not reached
        (0.1, 1, 1, 0.1, False, False),
        (0.1, 1, 1, 0.1, True, False),
        # cadence reached
        (0.1, 0.3, 0.1, 0.3, False, True),
        (0.1, 0.3, 0.1, 0.1, True, True),
        # model updated long back and
        # and cadence reached
        (0.1, 0.3, 0.1, 0.1, True, True),
    ]


@pytest.mark.parametrize(
    "log_frequency, current_log_step, last_log_step,"
    " last_model_update_step, check_model_update, expected",
    _log_ready_test_cases(),
)
def test_log_ready(
    log_frequency,
    current_log_step,
    last_log_step,
    last_model_update_step,
    check_model_update,
    expected,
):
    frequency_manager = FrequencyManager(log_frequency=log_frequency)
    frequency_manager.last_log_step = last_log_step
    frequency_manager.last_model_update_step = last_model_update_step

    actual = frequency_manager.log_ready(
        current_log_step=current_log_step, check_model_update=check_model_update
    )

    assert actual == expected


@pytest.mark.parametrize(
    "log_frequency, frequency_type, expectation",
    [
        # epoch frequency type accepts floats
        (0.1, "epoch", does_not_raise()),
        # batch and step frequency types need
        # log_frequency to be an integer
        (0.1, "step", pytest.raises(TypeError)),
        # negative log_frequency is invalid
        (-1, "epoch", pytest.raises(ValueError)),
        (-1, "step", pytest.raises(ValueError)),
    ],
)
def test__validate_log_frequency(log_frequency, frequency_type, expectation):
    with expectation:
        FrequencyManager(log_frequency=log_frequency, frequency_type=frequency_type)



import lm_eval
from compressed_tensors.quantization import (
    QuantizationArgs,
    QuantizationScheme,
    QuantizationStrategy,
    QuantizationType,
)
from lm_eval.utils import make_table
from transformers import AutoModelForCausalLM, AutoTokenizer

from llmcompressor import oneshot
from llmcompressor.modifiers.awq import AWQModifier
from llmcompressor.modifiers.quantization import QuantizationModifier

# This example demonstrates how to:
# 1) Run the `llm-compressor` implementation of AWQ
# 2) Evaluate the compressed model with the lm_eval framework

MODEL_ID = "meta-llama/Meta-Llama-3-8B-Instruct"
DATASET_ID = "mit-han-lab/pile-val-backup"
DATASET_SPLIT = "validation"
NUM_CALIBRATION_SAMPLES = 256
MAX_SEQUENCE_LENGTH = 512
OUTPUT_DIR = MODEL_ID.split("/")[-1] + "-awq-asym"

#
# 1) Run LLM Compressor AWQ implementation
#

recipe = [
    AWQModifier(bits=4, symmetric=False),
    QuantizationModifier(
        ignore=["lm_head"],
        config_groups={
            "group_0": QuantizationScheme(
                targets=["Linear"],
                weights=QuantizationArgs(
                    num_bits=4,
                    type=QuantizationType.INT,
                    dynamic=False,
                    symmetric=False,
                    strategy=QuantizationStrategy.GROUP,
                    group_size=128,
                ),
            )
        },
    ),
]

model = AutoModelForCausalLM.from_pretrained(
    MODEL_ID, device_map="auto", torch_dtype="auto"
)
tokenizer = AutoTokenizer.from_pretrained(MODEL_ID, trust_remote_code=True)


def get_calib_dataset(tokenizer):
    from datasets import load_dataset

    ds = load_dataset(
        DATASET_ID,
        split=f"{DATASET_SPLIT}[:{NUM_CALIBRATION_SAMPLES*100}]",
    )

    def preprocess(example):
        return {
            "input_ids": tokenizer.encode(example["text"].strip())[:MAX_SEQUENCE_LENGTH]
        }

    ds = (
        ds.shuffle(seed=42)
        .map(preprocess, remove_columns=ds.column_names)
        .filter(lambda example: len(example["input_ids"]) >= MAX_SEQUENCE_LENGTH)
        .select(range(NUM_CALIBRATION_SAMPLES))
    )

    return ds


oneshot(
    model=model,
    dataset=get_calib_dataset(tokenizer=tokenizer),
    recipe=recipe,
    output_dir=OUTPUT_DIR,
    max_seq_length=MAX_SEQUENCE_LENGTH,
    num_calibration_samples=NUM_CALIBRATION_SAMPLES,
)

print("Done! model saved to", OUTPUT_DIR)

#
# 2) Evaluate model on wikitext perplexity
#

results = lm_eval.simple_evaluate(
    model="vllm",
    model_args={
        "pretrained": OUTPUT_DIR,
        "add_bos_token": True,
        "dtype": "bfloat16",
        "gpu_memory_utilization": 0.5,
    },
    tasks=["wikitext"],
    num_fewshot=5,
    batch_size="auto",
)
print(make_table(results))



import json
import os
from typing import Any, Dict, List, Optional, Union

import torch
from loguru import logger
from safetensors import safe_open
from torch.nn import Module
from transformers import PreTrainedModel

from llmcompressor.core import active_session
from llmcompressor.typing import Processor

COMPLETED_STAGES_FILENAME = "completed_stages.json"

__all__ = [
    "copy_python_files_from_model_cache",
    "fallback_to_cpu",
    "parse_dtype",
    "get_session_model",
    "get_completed_stages",
    "save_completed_stages",
    "save_checkpoint",
]


def save_checkpoint(
    save_path: str,
    model: PreTrainedModel,
    processor: Optional[Processor] = None,
    save_safetensors: bool = True,
    save_compressed: bool = True,
    skip_sparsity_compression_stats: bool = False,
):
    """
    Save a model, processor, and recipe

    :param save_path: Path used to save model and processor
    :param model: model to save
    :param processor: processor to save
    :param save_safetensors: save model checkpoint using safetensors file type
    :param save_compressed: save model checkpoint using compressed-tensors format
    """
    # saving the model also saves the recipe
    model.save_pretrained(
        save_path,
        save_safetensors=save_safetensors,
        save_compressed=save_compressed,
        skip_sparsity_compression_stats=skip_sparsity_compression_stats,
    )
    if processor is not None:
        processor.save_pretrained(save_path)


def fallback_to_cpu(device: str) -> str:
    """
    Takes in a device string and forces it to cpu if cuda is not available

    :param device: device id to check
    :return: device modified for CUDA status
    """
    if "cuda" in device and not torch.cuda.is_available():
        logger.warning(
            f"Requested {device} but CUDA is not available, falling back to CPU"
        )
        return "cpu"

    return device


def parse_dtype(dtype_arg: Union[str, torch.dtype]) -> torch.dtype:
    """
    :param dtype_arg: dtype or string to parse
    :return: torch.dtype parsed from input string
    """
    dtype_arg = str(dtype_arg)
    dtype = "auto"  # get precision from model by default
    if dtype_arg in ("half", "float16", "torch.float16"):
        dtype = torch.float16
    elif dtype_arg in ("torch.bfloat16", "bfloat16"):
        dtype = torch.bfloat16
    elif dtype_arg in ("full", "float32", "torch.float32"):
        dtype = torch.float32

    return dtype


def get_session_model() -> Optional[Module]:
    """
    :return: pytorch module stored by the active CompressionSession,
        or None if no session is active
    """
    session = active_session()
    if not session:
        return None

    active_model = session.state.model
    return active_model


def get_completed_stages(checkpoint_dir: Any) -> List[str]:
    """
    Given a checkpoint directory for a staged run, get the list of stages that
    have completed in a prior run if the checkpoint_dir is a string

    :param checkpoint_dir: path to staged checkpoint
    :return: list of completed stage names
    """
    if isinstance(checkpoint_dir, str):
        stage_path = os.path.join(checkpoint_dir, COMPLETED_STAGES_FILENAME)
        if os.path.exists(stage_path):
            with open(stage_path) as stage_file:
                stage_data = json.load(stage_file)
                return stage_data["completed"]

    return []


def save_completed_stages(checkpoint_dir: str, completed_stages: List[str]):
    """
    Save a list of completed stages to a checkpoint directory

    :param checkpoint_dir: model checkpoint directory to save stages to
    :param completed_stages: list of stage names that have been run
    """
    stage_path = os.path.join(checkpoint_dir, COMPLETED_STAGES_FILENAME)
    with open(stage_path, "w") as out_file:
        json.dump({"completed": completed_stages}, out_file)


def load_safetensors_state_dict(file_path: str) -> Dict[str, torch.Tensor]:
    """
    Load a safetensors file from disk

    :param file_path: path to the safetensors file
    :return: dictionary of safetensors data
    """
    with safe_open(file_path, framework="pt", device="cpu") as f:
        return {key: f.get_tensor(key) for key in f.keys()}


def copy_python_files_from_model_cache(model, save_path: str):
    config = model.config
    cache_path = None
    if hasattr(config, "_name_or_path"):
        import os
        import shutil

        from huggingface_hub import hf_hub_download
        from transformers import TRANSFORMERS_CACHE
        from transformers.utils import http_user_agent

        cache_path = config._name_or_path
        if not os.path.exists(cache_path):
            user_agent = http_user_agent()
            config_file_path = hf_hub_download(
                repo_id=cache_path,
                filename="config.json",
                cache_dir=TRANSFORMERS_CACHE,
                force_download=False,
                user_agent=user_agent,
            )
            cache_path = os.path.sep.join(config_file_path.split(os.path.sep)[:-1])

        for file in os.listdir(cache_path):
            full_file_name = os.path.join(cache_path, file)
            if file.endswith(".py") and os.path.isfile(full_file_name):
                logger.debug(f"Transferring {full_file_name} to {save_path}")
                shutil.copy(full_file_name, save_path)



import pytest

from llmcompressor import LoggerConfig, configure_logger, logger


@pytest.fixture(autouse=True)
def reset_logger():
    # Ensure logger is reset before each test
    logger.remove()
    yield
    logger.remove()


def test_default_logger_settings(capsys):
    configure_logger()

    # Default settings should log to console with INFO level and no file logging
    logger.info("Info message")
    logger.debug("Debug message")

    captured = capsys.readouterr()
    assert captured.out.count("Info message") == 1
    assert "Debug message" not in captured.out


def test_configure_logger_console_settings(capsys):
    # Test configuring the logger to change console log level
    config = LoggerConfig(console_log_level="DEBUG")
    configure_logger(config=config)
    logger.info("Info message")
    logger.debug("Debug message")

    captured = capsys.readouterr()
    assert captured.out.count("Info message") == 1
    assert captured.out.count("Debug message") == 1


def test_configure_logger_file_settings(tmp_path):
    # Test configuring the logger to log to a file
    log_file = tmp_path / "test.log"
    config = LoggerConfig(log_file=str(log_file), log_file_level="DEBUG")
    configure_logger(config=config)
    logger.info("Info message")
    logger.debug("Debug message")

    with open(log_file, "r") as f:
        log_contents = f.read()
    assert log_contents.count('"message": "Info message"') == 1
    assert log_contents.count('"message": "Debug message"') == 1


def test_configure_logger_console_and_file(capsys, tmp_path):
    # Test configuring the logger to change both console and file settings
    log_file = tmp_path / "test.log"
    config = LoggerConfig(
        console_log_level="ERROR", log_file=str(log_file), log_file_level="INFO"
    )
    configure_logger(config=config)
    logger.info("Info message")
    logger.error("Error message")

    captured = capsys.readouterr()
    assert "Info message" not in captured.out
    assert captured.out.count("Error message") == 1

    with open(log_file, "r") as f:
        log_contents = f.read()
    assert log_contents.count('"message": "Info message"') == 1
    assert log_contents.count('"message": "Error message"') == 1


def test_environment_variable_override(monkeypatch, capsys, tmp_path):
    # Test environment variables override settings
    monkeypatch.setenv("LLM_COMPRESSOR_LOG_LEVEL", "ERROR")
    monkeypatch.setenv("LLM_COMPRESSOR_LOG_FILE", str(tmp_path / "env_test.log"))
    monkeypatch.setenv("LLM_COMPRESSOR_LOG_FILE_LEVEL", "DEBUG")

    configure_logger(config=LoggerConfig())
    logger.info("Info message")
    logger.error("Error message")
    logger.debug("Debug message")

    captured = capsys.readouterr()
    assert "Info message" not in captured.out
    assert captured.out.count("Error message") == 1
    assert "Debug message" not in captured.out

    with open(tmp_path / "env_test.log", "r") as f:
        log_contents = f.read()
    assert log_contents.count('"message": "Error message"') == 1
    assert log_contents.count('"message": "Info message"') == 1
    assert log_contents.count('"message": "Debug message"') == 1


def test_environment_variable_disable_logging(monkeypatch, capsys):
    # Test environment variable to disable logging
    monkeypatch.setenv("LLM_COMPRESSOR_LOG_DISABLED", "true")

    configure_logger(config=LoggerConfig())
    logger.info("Info message")
    logger.error("Error message")

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == ""


def test_environment_variable_enable_logging(monkeypatch, capsys):
    # Test environment variable to enable logging
    monkeypatch.setenv("LLM_COMPRESSOR_LOG_DISABLED", "false")

    configure_logger(config=LoggerConfig())
    logger.info("Info message")
    logger.error("Error message")

    captured = capsys.readouterr()
    assert captured.out.count("Info message") == 1
    assert captured.out.count("Error message") == 1



from transformers import Trainer as HFTransformersTrainer

from llmcompressor.transformers.finetune.session_mixin import SessionManagerMixIn

__all__ = ["Trainer"]


class Trainer(SessionManagerMixIn, HFTransformersTrainer):
    pass



from typing import Any, Optional, Tuple

import torch
from compressed_tensors.quantization.quant_args import QuantizationArgs
from compressed_tensors.quantization.utils import calculate_qparams
from torch import FloatTensor, IntTensor, Tensor

from llmcompressor.observers.base import Observer

__all__ = ["MovingAverageMSEObserver"]


@Observer.register("mse")
class MovingAverageMSEObserver(Observer):
    """
    Implements a dynamic quantization observer that sets the scale and
    zero point based on a moving average of the mse-clipped min and max observed values
    """

    def __init__(
        self,
        quantization_args: QuantizationArgs,
        averaging_constant: float = 0.01,
        grid: float = 100.0,
        maxshrink: float = 0.80,
        norm: float = 2.4,
    ):
        super().__init__(quantization_args=quantization_args)

        self.min_val = {}
        self.max_val = {}
        self.averaging_constant = averaging_constant
        self.grid = grid
        self.maxshrink = maxshrink
        self.norm = norm

    def calculate_mse_min_max(
        self,
        observed: Tensor,
        reduce_dims: Optional[Tuple[int]] = None,
    ):
        """
        Computes the mse-clipped min and max values of the observed tensor by
        optimizing for quantization error

        :param observed: observed tensor to calculate quantization parameters for
        :param reduce_dims: optional tuple of dimensions to reduce along,
            returned values will be shaped (1,) along the reduced dimensions
        :return: tuple of min and max values derived from the observed tensor
        """
        from compressed_tensors.quantization.lifecycle import fake_quantize

        if not reduce_dims:
            absolute_min_val, absolute_max_val = torch.aminmax(observed)
        else:
            absolute_min_val = torch.amin(observed, dim=reduce_dims, keepdims=True)
            absolute_max_val = torch.amax(observed, dim=reduce_dims, keepdims=True)

        best = torch.full_like(
            absolute_min_val, torch.finfo(absolute_min_val.dtype).max
        )
        min_val = torch.ones_like(absolute_min_val)
        max_val = torch.zeros_like(absolute_max_val)
        for i in range(int(self.maxshrink * self.grid)):
            p = 1 - i / self.grid
            shrinked_min_val = p * absolute_min_val
            shrinked_max_val = p * absolute_max_val

            candidate_scales, candidate_zero_points = calculate_qparams(
                shrinked_min_val, shrinked_max_val, self.quantization_args
            )
            q = fake_quantize(
                observed,
                candidate_scales,
                candidate_zero_points,
                self.quantization_args,
            )

            q -= observed
            q.abs_()
            q.pow_(self.norm)
            if not reduce_dims:
                err = torch.sum(q)
            else:
                err = torch.sum(q, reduce_dims, keepdims=True)

            tmp = err < best
            if torch.any(tmp):
                best[tmp] = err[tmp]
                min_val[tmp] = shrinked_min_val[tmp]
                max_val[tmp] = shrinked_max_val[tmp]
        return min_val, max_val

    def calculate_qparams(
        self,
        observed: Tensor,
        reduce_dims: Optional[Tuple[int]] = None,
        tensor_id: Optional[Any] = None,
    ) -> Tuple[FloatTensor, IntTensor]:
        """
        Updates the mse-clipped min and max values of the observed tensor using
        a moving average smoothed by the averaging_constant

        :param observed: observed tensor to calculate quantization parameters for
        :param reduce_dims: optional tuple of dimensions to reduce along,
            returned scale and zero point will be shaped (1,) along the
            reduced dimensions
        :param tensor_id: Optional id if different ranges of observed tensors are
            passed, useful for sharding tensors by group_size
        :return: tuple of scale and zero point derived from the observed tensor
        """
        min_val, max_val = self.calculate_mse_min_max(observed, reduce_dims)

        running_min_val = self.min_val.get(tensor_id, None)
        running_max_val = self.max_val.get(tensor_id, None)

        if running_min_val is None or running_max_val is None:
            updated_min_val = min_val
            updated_max_val = max_val
        else:
            updated_min_val = running_min_val + self.averaging_constant * (
                min_val - running_min_val
            )
            updated_max_val = running_max_val + self.averaging_constant * (
                max_val - running_max_val
            )

        tensor_id = tensor_id or "default"
        self.min_val[tensor_id] = updated_min_val
        self.max_val[tensor_id] = updated_max_val

        return calculate_qparams(
            updated_min_val, updated_max_val, self.quantization_args
        )

    def get_qparams_along_dim(
        self, observed, dim: int, tensor_id: Optional[Any] = None
    ):
        reduce_dims = tuple(idx for idx in range(observed.ndim) if idx != dim)
        return self.calculate_qparams(
            observed, reduce_dims=reduce_dims, tensor_id=tensor_id
        )

    def reset(self):
        """
        Reset the state of the observer, including min and maximum values
        """
        super().reset()
        self.min_val = {}
        self.max_val = {}



import os
import re
import weakref
from functools import wraps
from typing import Dict, Optional

import torch
import transformers
from accelerate.accelerator import get_state_dict_offloaded_model
from compressed_tensors import (
    CompressionFormat,
    ModelCompressor,
    SparsityCompressionConfig,
    is_module_offloaded,
    update_offload_parameter,
)
from loguru import logger
from safetensors.torch import storage_ptr
from transformers import PreTrainedModel

from llmcompressor.core import active_session
from llmcompressor.pytorch.model_load.helpers import copy_python_files_from_model_cache
from llmcompressor.recipe.recipe import Recipe
from llmcompressor.transformers.compression.quantization_format import (
    infer_quantization_format,
)
from llmcompressor.transformers.compression.sparsity_metadata_config import (
    SparsityConfigMetadata,
)
from llmcompressor.transformers.utils import RECIPE_FILE_NAME
from llmcompressor.transformers.utils.helpers import infer_recipe_from_model_path

__all__ = ["modify_save_pretrained"]


def modify_save_pretrained(model: PreTrainedModel):
    """
    Overrides a PreTrainedModel's save_pretrained() method with a wrapped version that
    supports compression. The new save_pretrained function performs the following saving
    operations:

    1. Saves the model state, potentially in a compressed format
    2. Saves the recipe, appending any current recipes to existing recipe files
    3. Copies any necessary python files from the model cache
    """

    def save_pretrained_compressed(save_pretrained_method):
        if getattr(save_pretrained_method, "_overridden", False):
            # `model.save_pretrained` has already been replaced, return.
            return save_pretrained_method

        # Keep a weak reference to the model class and unbound save_pretrained
        # method so we can call the original
        model_ref = weakref.ref(save_pretrained_method.__self__)
        original_save_pretrained = save_pretrained_method.__func__
        model_class = model_ref().__class__
        del save_pretrained_method

        @wraps(original_save_pretrained)
        def save_pretrained_wrapper(
            save_directory: str,
            sparsity_config: Optional[SparsityCompressionConfig] = None,
            quantization_format: Optional[str] = None,
            save_compressed: bool = True,
            safe_serialization: bool = True,
            skip_sparsity_compression_stats: bool = True,
            disable_sparse_compression: bool = False,
            **kwargs,
        ):
            """
            Wrapper around PreTrainedModel.save_pretrained(), adds functionality for
            saving models in a compressed format on disk. The compression format is
            saved to the model's config file

            :param save_directory: output directory to save model to
            :param sparsity_config: optional sparsity config to compress model with,
                if no config is provided it will be inferred from the model
            :param quantization_format: optional compression format for quantized
                models. If none is provided it will be inferred from the model
            :param save_compressed: whether or not to compress the model on disk
            :param skip_sparsity_compression_stats: whether to skip the calculation of
                sparsity statistics (such as global sparsity and sparsity structure)
                when saving a model in dense format
            :param disable_sparse_compression: whether to skip sparse compression
                during save, default is False
            :param kwargs: additional kwargs to pass on to model.save_pretrained
            """

            # HACK: Override the dtype_byte_size function in transformers to
            # support float8 types. Fix is posted upstream
            # https://github.com/huggingface/transformers/pull/30488
            transformers.modeling_utils.dtype_byte_size = new_dtype_byte_size

            # state_dict gets passed in as a kwarg for FSDP models
            state_dict = kwargs.pop("state_dict", None)
            if state_dict is None:
                logger.info("Fetching state_dict - this may take some time")
                state_dict = get_state_dict_offloaded_model(model)

            logger.info("Fetching compressor")
            compressor = get_model_compressor(
                model=model,
                sparsity_config=sparsity_config,
                quantization_format=quantization_format,
                save_compressed=save_compressed,
                skip_sparsity_compression_stats=skip_sparsity_compression_stats,
                state_dict=state_dict,
                disable_sparse_compression=disable_sparse_compression,
            )

            if compressor is None:
                # model is not compressed or quantized, save as normal
                original_save_pretrained_func = original_save_pretrained.__get__(
                    model, model_class
                )
                original_save_pretrained_func(
                    save_directory, state_dict=state_dict, **kwargs
                )
                return

            # make sure we're on the main process when saving
            if state_dict is not None and len(state_dict) > 0:
                compressed_state_dict = compressor.compress(model, state_dict)
                logger.info("Saving compressed model to disk")
                original_save_pretrained.__get__(model, model_class)(
                    save_directory,
                    state_dict=compressed_state_dict,
                    safe_serialization=safe_serialization,
                    **kwargs,
                )
                compressor.update_config(save_directory)

            # update existing recipe
            update_and_save_recipe(model.name_or_path, save_directory)

            # copy python files from cache dir to save_path if any
            copy_python_files_from_model_cache(model, save_directory)

        save_pretrained_wrapper._overriden = True
        return save_pretrained_wrapper

    # wrap save_pretrained if not already
    if not getattr(model.save_pretrained, "_overriden", False):
        model.save_pretrained = save_pretrained_compressed(model.save_pretrained)


# HACK: Override the dtype_byte_size function in transformers to support float8 types
# Fix is posted upstream https://github.com/huggingface/transformers/pull/30488
def new_dtype_byte_size(dtype):
    if dtype == torch.bool:
        return 1 / 8
    bit_search = re.search(r"[^\d](\d+)_?", str(dtype))
    if bit_search is None:
        raise ValueError(f"`dtype` is not a valid dtype: {dtype}.")
    bit_size = int(bit_search.groups()[0])
    return bit_size // 8


def patch_tied_tensors_bug(model: torch.nn.Module):
    """
    Patches bug where HF transformers will fail to untie weights under specific
    circumstances (https://github.com/huggingface/transformers/issues/33689).

    This function detects those cases and unties the tensors if applicable

    :param model: model to fix
    """
    if (
        hasattr(model.config, "tie_word_embeddings")
        and not model.config.tie_word_embeddings
    ):
        input_embed = model.get_input_embeddings()
        output_embed = model.get_output_embeddings()

        if input_embed is None or output_embed is None:
            # some models fail to properly override the abstract methods
            return

        if storage_ptr(input_embed.weight) == storage_ptr(output_embed.weight):
            for module in (input_embed, output_embed):
                if not is_module_offloaded(module):
                    # create new storage ptr for onloaded weight
                    untied_data = module.weight.data.clone()
                    module.weight.data = untied_data
                else:
                    # create new storage ptr for offloaded weight
                    # note `update_offload_parameter` does not create a new storage ptr
                    untied_data = module._hf_hook.weights_map["weight"].clone()
                    update_offload_parameter(module, "weight", untied_data)


def get_model_compressor(
    model: torch.nn.Module,
    sparsity_config: Optional[SparsityCompressionConfig] = None,
    quantization_format: Optional[str] = None,
    save_compressed: bool = True,
    skip_sparsity_compression_stats: bool = True,
    state_dict: Optional[Dict] = None,
    disable_sparse_compression: bool = False,
):
    """
    Obtain the compressor based on the config and the
        quantization_format

    :param model: torch model
    :param sparsify_config: Sparsity Compression config
    :param quantization_format: Format that the model was quantized to.
        if not provivided, will be extrapolated from `infer_quantization_format`
    :param save_compressed: boolean representing to save in a compressed
        format
    :param skip_sparsity_compression_stats: bool allowing compression stats on std out
    :param state_dict: state_dict of the model
    :param disable_sparse_compression: bool to skip sparse compression
    """
    # find offloaded state dict if none is provided
    if state_dict is None:
        state_dict = get_state_dict_offloaded_model(model)

    if sparsity_config is None:
        """
        Case 1: No sparsity config is provided
            1. Will either skip sparsity compression
            2. Or we will infer sparsity from the model directly

        Check recipe for applied sparsity:
            - Set skip_sparsity_compression_stats to False if don't find a
                sparsity structure from the recipe
            - If we identify sparsity based on the recipe or the user
                set skip_sparsity_compression_stats to False, generate config
        """
        sparsity_structure = SparsityConfigMetadata.infer_sparsity_structure(
            model, check_only_modifiers=True
        )
        if sparsity_structure is not None:
            skip_sparsity_compression_stats = False

        if skip_sparsity_compression_stats:
            logger.info(
                "skip_sparsity_compression_stats set to True. Skipping sparsity "
                "compression statistic calculations. No sparsity compressor will "
                "be applied."
            )
            sparsity_config = None
        else:
            sparsity_config = SparsityConfigMetadata.from_pretrained(
                model,
                state_dict=state_dict,
                compress=save_compressed,
                quantization_format=quantization_format,
                disable_sparse_compression=disable_sparse_compression,
                sparsity_structure=sparsity_structure,
            )
    else:
        """
        # Case 2: User provides a Sparsity Config
            - This is the case when there is existing sparsity in the
                model that we'd like to account for while compressing
            - Users should provide a SparsityConfig, conveying the model's
                sparsity structure when saving the model
        """
        if sparsity_config.sparsity_structure is None:
            logger.info(
                "SparsityConfigMetadata provided without indicating ",
                "the sparsity structure. Sparisty will be inferred from the model. "
                "Consider providing the structure to skip this step ",
            )
            sparsity_config.sparsity_structure = (
                SparsityConfigMetadata.infer_sparsity_structure(model)
            )

    quantization_format: Optional[CompressionFormat] = infer_quantization_format(
        model=model,
        quantization_format=quantization_format,
        save_compressed=save_compressed,
        sparsity_structure=None
        if sparsity_config is None
        else sparsity_config.sparsity_structure,
    )

    return ModelCompressor.from_pretrained_model(
        model,
        sparsity_config=sparsity_config,
        quantization_format=quantization_format,
    )


def update_and_save_recipe(model_stub: str, save_directory: str):
    """
    Save a recipe ontop of any existing recipe files located at model_stub

    :param model_stub: path to existing model or model stub which may contain an
        existing recipe
    :param save_directory: path to save combined existing recipe and current recipe
    """
    recipes_to_save = []
    existing_recipe = infer_recipe_from_model_path(model_stub)
    if existing_recipe is not None:
        recipes_to_save.append(existing_recipe)

    new_recipe = active_session().lifecycle.recipe_container.compiled_recipe
    if new_recipe is not None:
        recipes_to_save.append(new_recipe)

    recipe = Recipe.simplify_combine_recipes(recipes_to_save)

    # save recipe
    recipe_path = os.path.join(save_directory, RECIPE_FILE_NAME)
    recipe.yaml(recipe_path)
